# ----------------------------------------------------------------------------
# This file is simply the entrypoint from the initial call in ida_binsync,
# which will setup all the hooks for both the UI and IDB changes, and will
# also create the config window.
#
# ----------------------------------------------------------------------------
import logging
import os

from PyQt5 import sip
from PyQt5.QtCore import QObject

import idaapi
import ida_kernwin
import idc
import ida_hexrays
import idautils
from PyQt5.QtWidgets import QWidget, QVBoxLayout

from binsync.common.ui.version import set_ui_version
set_ui_version("PyQt5")
from binsync.common.ui.config_dialog import SyncConfig
from binsync.common.ui.control_panel import ControlPanel
from binsync.common.ui.magic_sync_dialog import display_magic_sync_dialog

from .hooks import MasterHook
from . import IDA_DIR, VERSION
from .controller import IDABinSyncController

l = logging.getLogger(__name__)
controller = IDABinSyncController()

# disable the annoying "Running Python script" wait box that freezes IDA at times
idaapi.set_script_timeout(0)


#
#   UI Hook, placed here for convenience of reading UI implementation
#

class ScreenHook(ida_kernwin.View_Hooks):
    def __init__(self):
        super(ScreenHook, self).__init__()
        self.hooked = False

    def view_click(self, view, event):
        form_type = idaapi.get_widget_type(view)
        decomp_view = idaapi.get_widget_vdui(view)
        if not form_type:
            return

        # check if view is decomp or disassembly before doing expensive ea lookup
        if not decomp_view and not form_type == idaapi.BWN_DISASM:
            return

        ea = idc.get_screen_ea()
        if not ea:
            return

        controller.update_active_context(ea)
#
#   Action Handlers
#

class IDAActionHandler(idaapi.action_handler_t):
    def __init__(self, action, plugin, typ):
        super(IDAActionHandler, self).__init__()
        self.action = action
        self.plugin = plugin
        self.typ = typ

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS



#
# Control Panel
#

class ControlPanelViewWrapper(object):
    NAME = "BinSync"

    def __init__(self, controller):
        # create a dockable view
        self.twidget = idaapi.create_empty_widget(ControlPanelViewWrapper.NAME)
        self.widget = sip.wrapinstance(int(self.twidget), QWidget)
        self.widget.name = ControlPanelViewWrapper.NAME
        self.width_hint = 250

        self._controller = controller
        self._w = None

        self._init_widgets()

    def _init_widgets(self):
        self._w = ControlPanel(self._controller)
        layout = QVBoxLayout()
        layout.addWidget(self._w)
        layout.setContentsMargins(2,2,2,2)
        self.widget.setLayout(layout)

#
#   Base Plugin
#


class BinsyncPlugin(QObject, idaapi.plugin_t):
    """Plugin entry point. Does most of the skinning magic."""

    flags = idaapi.PLUGIN_FIX
    comment = "Syncing dbs between users"

    help = "This is help"
    wanted_name = "Binsync: settings"
    wanted_hotkey = "Ctrl-Shift-B"

    def __init__(self, *args, **kwargs):
        print("[BinSync] {} loaded!".format(VERSION))

        QObject.__init__(self, *args, **kwargs)
        idaapi.plugin_t.__init__(self)
        self.hooks_started = False

    def open_config_dialog(self):
        dialog = SyncConfig(controller)
        dialog.exec_()

        if not controller.check_client():
            return

        if not self.hooks_started:
            self.action_hooks.hook()
            self.view_hook.hook()

        self.open_control_panel()

        if dialog.open_magic_sync:
            #display_magic_sync_dialog(controller)
            l.debug("Magic Sync is disabled on startup for now.")

    def open_control_panel(self):
        """
        Open the control panel view and attach it to IDA View-A or Pseudocode-A.
        """
        wrapper = ControlPanelViewWrapper(controller)
        if not wrapper.twidget:
            l.info("BinSync is unable to find a widget to attach to. You are likely running headlessly")
            return None

        flags = idaapi.PluginForm.WOPN_TAB | idaapi.PluginForm.WOPN_RESTORE | idaapi.PluginForm.WOPN_PERSIST
        idaapi.display_widget(wrapper.twidget, flags)
        wrapper.widget.visible = True

        # casually open a pseudocode window, this prevents magic sync from spawning pseudocode windows
        # in weird locations upon an initial run
        func_addr = next(idautils.Functions())
        if controller.decompiler_available:
            ida_hexrays.open_pseudocode(func_addr, ida_hexrays.OPF_NO_WAIT | ida_hexrays.OPF_REUSE)

        # then attempt to flip back to IDA View-A
        twidget = idaapi.find_widget("IDA View-A")
        if twidget is not None:
            ida_kernwin.activate_widget(twidget, True)

        target = "Functions"
        fwidget = idaapi.find_widget(target)

        if not fwidget:
            # prioritize attaching the binsync panel to a decompilation window
            target = "Pseudocode-A"
            dwidget = idaapi.find_widget(target)

            if not dwidget:
                target = "IDA View-A"

        if target == "Functions":
            idaapi.set_dock_pos(ControlPanelViewWrapper.NAME, target, idaapi.DP_INSIDE)
        else:
            # attach the panel to the found target
            idaapi.set_dock_pos(ControlPanelViewWrapper.NAME, target, idaapi.DP_RIGHT)

    def install_actions(self):
        self.install_control_panel_action()

    def install_control_panel_action(self):
        action_id = "binsync:control_panel"
        action_desc = idaapi.action_desc_t(
            action_id,
            "BinSync: ~C~ontrol Panel",
            IDAActionHandler(self.open_control_panel, None, None),
            None,
            "Open the BinSync control panel",
        )
        result = idaapi.register_action(action_desc)
        if not result:
            l.info("BinSync is unable to find a widget to attach to. You are likely running headlessly")
            return None

        result = idaapi.attach_action_to_menu(
            "View/Open subviews/Hex dump",
            action_id,
            idaapi.SETMENU_INS,
        )
        if not result:
            l.info("BinSync is unable to find a widget to attach to. You are likely running headlessly")
            return None

    def _init_hooks(self):
        # Hook UI Startup in IDA
        self.install_actions()

        # init later
        self.view_hook = ScreenHook()
        # Hook IDB & Decomp Actions in IDA
        self.action_hooks = MasterHook(controller)

    def init(self):
        self._init_hooks()

        return idaapi.PLUGIN_KEEP

    def run(self, arg):
        self.open_config_dialog()

    def term(self):
        print("term() called!")

#
#   Utils
#


def plugin_resource(resource_name):
    """
    Return the full path for a given plugin resource file.
    """
    plugin_path = os.path.abspath(os.path.dirname(__file__))

    return os.path.join(
        plugin_path,
        resource_name
    )







# ----------------------------------------------------------------------------
# This file contains the BinSyncController class which acts as the the
# bridge between the plugin UI and direct calls to the binsync client found in
# the core of binsync. In the controller, you will find code used to make
# pushes and pulls of user changes.
#
# You will also notice that the BinSyncController runs two extra threads in
# it:
#   1. BinSync "git pulling" thread to constantly get changes from others
#   2. Command Routine to get hooked changes to IDA attributes
#
# The second point is more complicated because it acts as the queue of
# runnable actions that are queued from inside the hooks.py file.
# Essentially, every change that happens in IDA from the main user triggers
# a hook which will push an action to be preformed onto the command queue;
# Causing a "git push" on every change.
#
# ----------------------------------------------------------------------------

from functools import wraps
import re
import threading
import time
import datetime
import logging
from typing import Dict, List, Tuple, Optional
from collections import OrderedDict, defaultdict

import idc
import idaapi
import idautils
import ida_struct
import ida_hexrays
import ida_funcs
import ida_kernwin

import binsync
from binsync.common.controller import BinSyncController, init_checker, fill_event
from binsync import (
    StackVariable, StackOffsetType, Function, FunctionHeader, Struct, Comment, GlobalVariable, Enum, State, Patch
)
from . import compat
from .artifact_lifter import IDAArtifactLifter

_l = logging.getLogger(name=__name__)


def update_on_view(f):
    wraps(f)
    def _update_on_view(self: IDABinSyncController, func_addr, *args, **kwargs):
        # always execute something we are looking at
        active_ctx = self.active_context()
        if active_ctx and active_ctx.addr == func_addr:
            return f(self, func_addr, *args, **kwargs)

        # otherwise, execute next time we look at it through the hooks
        task = UpdateTask(f, self, func_addr, *args, **kwargs)
        self.update_states[func_addr].add_update_task(task)

    return _update_on_view


#
#   Wrapper Classes
#


class UpdateTask:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __eq__(self, other):
        return (
            isinstance(other, UpdateTask)
            and other.func == other.func
            and other.kwargs == other.kwargs
        )

    def __hash__(self):
        expanded_kwargs = list()
        for k, v in self.kwargs.items():
            expanded_kwargs.append(f"{k}={v}")
        return hash((self.func, *self.args, *expanded_kwargs))

    def dump(self):
        return self.func, self.args, self.kwargs


class UpdateTaskState:
    def __init__(self):
        self.update_tasks: Dict[UpdateTask, bool] = OrderedDict()
        self.update_tasks_lock = threading.Lock()

    def toggle_auto_sync_task(self, update_task):
        with self.update_tasks_lock:
            # delete the task if it is an auto_sync task already
            if update_task in list(self.update_tasks.keys()) and self.update_tasks[update_task]:
                del self.update_tasks[update_task]
                return False

            # set/make the task if its not auto_sync already
            self.update_tasks[update_task] = True
            return True

    def add_update_task(self, update_task):
        with self.update_tasks_lock:
            self.update_tasks[update_task] = False

    def do_updates(self):
        with self.update_tasks_lock:
            # run each task in the update task queue
            for update_task in list(self.update_tasks.keys()):
                f, args, kwargs = update_task.dump()
                auto_sync = self.update_tasks[update_task]

                # doing task
                try:
                    f(*args, **kwargs)
                except Exception:
                    print(f"[BinSync]: failed to execute cache of {f} with {args}")

                # remove the task if its not an auto_sync task
                if not auto_sync:
                    del self.update_tasks[update_task]


#
#   Controller
#

class IDABinSyncController(BinSyncController):
    def __init__(self):
        super(IDABinSyncController, self).__init__(artifact_lifter=IDAArtifactLifter(self))

        # view change callback
        self._updated_ctx = None
        self._decompiler_available = None

        # update state for only updating when needed
        self.update_states = defaultdict(UpdateTaskState)

    #
    # Controller Interaction
    #

    def binary_hash(self) -> str:
        return idc.retrieve_input_file_md5().hex()

    def active_context(self):
        return self._updated_ctx

    def update_active_context(self, addr):
        if not addr or addr == idaapi.BADADDR:
            return

        func_addr = compat.ida_func_addr(addr)
        if func_addr is None:
            return

        func = binsync.data.Function(
            func_addr, 0, header=FunctionHeader(compat.get_func_name(func_addr), func_addr)
        )
        self._updated_ctx = func

    def binary_path(self) -> Optional[str]:
        return compat.get_binary_path()

    def get_func_size(self, func_addr) -> int:
        return compat.get_func_size(func_addr)

    def goto_address(self, func_addr) -> None:
        compat.jumpto(func_addr)

    @property
    def decompiler_available(self) -> bool:
        if self._decompiler_available is None:
            self._decompiler_available = ida_hexrays.init_hexrays_plugin()

        return self._decompiler_available

    #
    # IDA DataBase Fillers
    #

    @init_checker
    @fill_event
    def fill_struct(self, struct_name, user=None, header=True, members=True, artifact=None, **kwargs):
        data_changed = False
        struct: Struct = artifact

        if not struct:
            _l.warning("Was not able to find the struct you requested in other users name")
            return None

        if header:
            data_changed |= compat.set_ida_struct(struct, self)

        if members:
            data_changed |= compat.set_ida_struct_member_types(struct, self)

        return data_changed

    @init_checker
    @fill_event
    def fill_global_var(self, var_addr, user=None, artifact=None, **kwargs):
        changed = False
        global_var: GlobalVariable = artifact
        if global_var and global_var.name:
            changed = compat.set_global_var_name(var_addr, global_var.name)

        if changed:
            ctx = self.active_context()
            if ctx:
                compat.refresh_pseudocode_view(ctx.addr)

        return changed

    @init_checker
    @fill_event
    def fill_function(self, func_addr, user=None, artifact=None, **kwargs):
        func: Function = artifact
        try:
            ida_code_view = compat.acquire_pseudocode_vdui(func.addr)
        except Exception:
            ida_code_view = None

        changes = super(IDABinSyncController, self).fill_function(
            func_addr, user=user, artifact=artifact, ida_code_view=ida_code_view, **kwargs
        )
        compat.refresh_pseudocode_view(func.addr)
        return changes

    @init_checker
    @fill_event
    def fill_comment(self, addr, user=None, artifact=None, **kwargs):
        cmt: Comment = artifact
        res = compat.set_ida_comment(addr, cmt.comment, decompiled=cmt.decompiled)
        if not res:
            _l.warning(f"Failed to sync comment at <{hex(addr)}>: \'{cmt.comment}\'")

        return res

    @init_checker
    @fill_event
    def fill_stack_variable(self, func_addr, offset, user=None, artifact=None, ida_code_view=None, **kwargs):
        if ida_code_view is None:
            return False

        stack_var: StackVariable = artifact
        frame = idaapi.get_frame(stack_var.addr)
        changes = False
        if frame is None or frame.memqty <= 0:
            _l.warning(f"Function {stack_var.addr:x} does not have an associated function frame. Stopping sync here!")
            return False

        if ida_struct.set_member_name(frame, offset, stack_var.name):
            changes |= True

        ida_type = compat.convert_type_str_to_ida_type(stack_var.type)
        if ida_type is None:
            _l.warning(f"IDA Failed to parse type for stack var {stack_var}")
            return changes

        changes |= compat.set_stack_vars_types({offset: ida_type}, ida_code_view, self)
        return changes

    @init_checker
    @fill_event
    def fill_function_header(self, func_addr, user=None, artifact=None, ida_code_view=None, **kwargs):
        func_header: FunctionHeader = artifact
        if ida_code_view is None:
            compat.set_ida_func_name(func_addr, func_header.name)
            return False

        updated_header = compat.set_function_header(ida_code_view, func_header)
        return updated_header

    @init_checker
    @fill_event
    def fill_enum(self, name, user=None, artifact=None, ida_code_view=None, **kwargs):
        enum: Enum = artifact
        updated_enum = compat.set_enum(enum)
        return updated_enum

    #
    # Artifact API
    #

    def functions(self) -> Dict[int, Function]:
        return compat.functions()

    def function(self, addr) -> Optional[Function]:
        return compat.function(addr, decompiler_available=self.decompiler_available)

    def global_vars(self) -> Dict[int, GlobalVariable]:
        return compat.global_vars()

    def global_var(self, addr) -> Optional[GlobalVariable]:
        return compat.global_var(addr)

    def structs(self) -> Dict[str, Struct]:
        return compat.structs()

    def struct(self, name) -> Optional[Struct]:
        return compat.struct(name)

    def enums(self) -> Dict[str, Enum]:
        return compat.enums()

    def enum(self, name) -> Optional[Enum]:
        return compat.enum(name)



from typing import Optional
import logging

from binsync.common.controller import BinSyncController, init_checker, fill_event
from binsync.data import (
    Function, FunctionHeader, StackVariable, Comment
)

from .ghidra_client import BSGhidraClient
from .artifact_lifter import GhidraArtifactLifter

l = logging.getLogger(__name__)


class GhidraBinSyncController(BinSyncController):
    def __init__(self):
        super(GhidraBinSyncController, self).__init__(GhidraArtifactLifter(self))
        self.ghidra = BSGhidraClient()
        self.base_addr = None

    def binary_hash(self) -> str:
        return self.ghidra.binary_hash

    def active_context(self):
        return self.ghidra.context()

    def binary_path(self) -> Optional[str]:
        return self.ghidra.binary_path

    def get_func_size(self, func_addr) -> int:
        return 0

    def rebase_addr(self, addr, up=True):
        if self.base_addr is None:
            self.base_addr = self.ghidra.base_addr

        if up:
            if addr > self.base_addr:
                return
            return addr + self.base_addr
        elif addr > self.base_addr:
            return addr - self.base_addr

    def goto_address(self, func_addr) -> None:
        self.ghidra.goto_address(func_addr)

    #
    # Ghidra Specific
    #

    def connect_ghidra_client(self):
        return self.ghidra.connect()

    def alert_ghidra_ui_configured(self):
        status = True if self.check_client() else False
        self.ghidra.alert_ui_configured(status)

    #
    # BinSync API
    #

    @fill_event
    def fill_stack_variable(self, func_addr, offset, user=None, artifact=None, **kwargs):
        update = False
        stack_var: StackVariable = artifact
        if stack_var.name:
            update |= self.ghidra.set_stack_var_name(stack_var.addr, stack_var.stack_offset, stack_var.name)

        if stack_var.type:
            update |= self.ghidra.set_stack_var_type(stack_var.addr, stack_var.stack_offset, stack_var.type)

        return update

    @fill_event
    def fill_function_header(self, func_addr, user=None, artifact=None, **kwargs):
        update = False
        func_header: FunctionHeader = artifact
        if func_header.name:
            update |= self.ghidra.set_func_name(func_header.addr, func_header.name)

        if func_header.ret_type:
            update |= self.ghidra.set_func_rettype(func_header.addr, func_header.ret_type)

        return update

    @fill_event
    def fill_comment(self, addr, user=None, artifact=None, **kwargs):
        update = False
        comment: Comment = artifact
        if comment.comment:
            update |= self.ghidra.set_comment(comment.addr, comment.comment, comment.decompiled)

        return update

    @init_checker
    def magic_fill(self, preference_user=None):
        super(GhidraBinSyncController, self).magic_fill(
            preference_user=preference_user, target_artifacts={Function: self.fill_function}
        )



from binsync.common import ArtifactLifter


class AngrArtifactLifter(ArtifactLifter):
    """
    TODO: fix me
    """
    def __init__(self, controller):
        super(AngrArtifactLifter, self).__init__(controller)

    def lift_addr(self, addr: int) -> int:
        return self.controller.rebase_addr(addr)

    def lift_type(self, type_str: str) -> str:
        return type_str

    def lift_stack_offset(self, offset: int, func_addr: int) -> int:
        return offset

    def lower_addr(self, addr: int) -> int:
        return self.controller.rebase_addr(addr, up=True)

    def lower_type(self, type_str: str) -> str:
        return type_str

    def lower_stack_offset(self, offset: int, func_addr: int) -> int:
        return offset



import toml

from binsync.data.artifact import Artifact


class Comment(Artifact):
    __slots__ = (
        "last_change",
        "addr",
        "func_addr",
        "comment",
        "decompiled",

    )

    def __init__(self, addr, comment,  func_addr=None, decompiled=False, last_change=None):
        super(Comment, self).__init__(last_change=last_change)
        self.comment = comment  # type: str
        self.decompiled = decompiled  # TODO: use this in other places!
        self.addr = addr  # type: int
        self.func_addr = func_addr

    def __str__(self):
        return f"<Comment: @{hex(self.addr)} len={len(self.comment)}>"

    def __repr__(self):
        return self.__str__()

    @classmethod
    def parse(cls, s):
        comm = Comment(None, None)
        comm.__setstate__(toml.loads(s))
        return comm

    @classmethod
    def load_many(cls, comms_toml):
        for comm_toml in comms_toml.values():
            comm = Comment(None, None)
            try:
                comm.__setstate__(comm_toml)
            except TypeError:
                # skip all incorrect ones
                continue
            yield comm

    @classmethod
    def dump_many(cls, comments):
        comments_ = {}

        for v in sorted(comments.values(), key=lambda x: x.addr):
            comments_["%x" % v.addr] = v.__getstate__()
        return comments_

    def copy(self):
        return Comment(
            self.addr,
            self.comment,
            func_addr=self.func_addr,
            decompiled=self.decompiled,
            last_change=self.last_change
        )

    def nonconflict_merge(self, obj2: "Comment", **kwargs) -> "Comment":
        obj1: "Comment" = self.copy()
        if not obj2 or obj1 == obj2:
            return obj1

        merge_comment = obj1
        return merge_comment



import logging

from binsync.common.ui.force_push.panels.functions_table import QFunctionTable
from binsync.common.ui.force_push.panels.global_panel import QGlobalsTable
from binsync.common.ui.qt_objects import (
    QTabWidget,
    QVBoxLayout,
    QWidget,
    Signal
)

l = logging.getLogger(__name__)

class ForcePushUI(QWidget):
    update_ready = Signal()
    def __init__(self, controller, parent=None):
        super(ForcePushUI, self).__init__(parent)
        self.controller = controller

        self.tables = {}
        self._init_widgets()
        self._update_table_data()

    def _init_widgets(self):
        # tabs for panel_tabs
        self.tabView = QTabWidget()
        self.tabView.setContentsMargins(0, 0, 0, 0)

        # add panel_tabs to tabs
        self._func_table = QFunctionTable(self.controller)
        self._global_table = QGlobalsTable(self.controller)

        self.tabView.addTab(self._func_table, "Functions")
        self.tabView.addTab(self._global_table, "Globals")

        self.tables.update({
            "functions": self._func_table,
            "globals": self._global_table
        })

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.tabView)
        main_layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(main_layout)

    def _update_table_data(self):
        for _, table in self.tables.items():
            table.update_table()



from collections import defaultdict
from threading import Lock
import logging

l = logging.getLogger(__name__)


class Cache:
    def __init__(self):
        self.state_cache = defaultdict(StateCache)
        self.user_cache = UserCache()
        self.state_lock = Lock()
        self.user_lock = Lock()

    def update_state_cache_commits(self, username_commit_dict: dict):
        for username, commit in username_commit_dict.items():
            with self.state_lock:
                if self.state_cache[username].commit != commit:
                    self.state_cache[username].state = None
                    self.state_cache[username].commit = commit

    def update_user_cache_branches(self, branch_set: set):
        with self.user_lock:
            if branch_set != self.user_cache.known_branches:
                self.user_cache.users = None
                self.user_cache.known_branches = branch_set

    #
    # getters
    #

    def get_state(self, user=None, **kwargs):
        with self.state_lock:
            state = self.state_cache[user].state
            return state.copy() if state else state

    def users(self, **kwargs):
        with self.user_lock:
            return self.user_cache.users if self.user_cache.users else None

    #
    # setters
    #

    def set_state(self, state, user=None, **kwargs):
        with self.state_lock:
            self.state_cache[user].state = state

    def set_users(self, users, **kwargs):
        with self.user_lock:
            self.user_cache.users = users


class StateCache:
    def __init__(self, state=None, commit=None):
        self.state = state
        self.commit = commit


class UserCache:
    def __init__(self, users=None):
        self.users = users
        self.known_branches = set()


import logging

from binsync.common.controller import BinSyncController
from binsync.common.ui.qt_objects import (
    QAbstractItemView,
    QHeaderView,
    QMenu,
    Qt,
    QTableWidget,
    QTableWidgetItem,
)
from binsync.common.ui.utils import QNumericItem, friendly_datetime
from binsync.data import State

l = logging.getLogger(__name__)

class QCTXItem:
    """
    The CTX view shown in the Control Panel. Responsible for showing the main user info on whatever the main user
    is currently looking at (clicked). For any line in a function, this would be the entire function. For a struct,
    this would be a struct. The view will be as useful as the decompilers support for understanding what the user
    is looking at.

    TODO: refactor this to allow for any context item, not just functions (like structs).
    """
    def __init__(self, user, name, last_push, changes):
        self.user = user
        self.name = name
        self.last_push = last_push
        self.changes = changes

    def widgets(self):
        user = QTableWidgetItem(self.user)
        name = QTableWidgetItem(self.name)

        # sort by unix value
        last_push = QNumericItem(friendly_datetime(self.last_push))
        last_push.setData(Qt.UserRole, self.last_push)

        changes = QNumericItem(self.changes)
        changes.setData(Qt.UserRole, self.changes)

        widgets = [
            user,
            name,
            last_push,
            changes
        ]

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QCTXTable(QTableWidget):

    HEADER = [
        'User',
        'Remote Name',
        'Last Push',
        'Changes'
    ]

    def __init__(self, controller: BinSyncController, parent=None):
        super(QCTXTable, self).__init__(parent)
        self.controller = controller
        self.items = []
        self.ctx = None

        # header
        self.setColumnCount(len(self.HEADER))
        self.setHorizontalHeaderLabels(self.HEADER)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.horizontalHeader().setHorizontalScrollMode(self.ScrollPerPixel)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignHCenter | Qt.Alignment(Qt.TextWordWrap))
        self.horizontalHeader().setMinimumWidth(160)
        self.setHorizontalScrollMode(self.ScrollPerPixel)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(24)

        self.setSortingEnabled(True)

    def reload(self):
        self.setSortingEnabled(False)
        self.setRowCount(len(self.items))

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        self.viewport().update()
        self.setSortingEnabled(True)

    def contextMenuEvent(self, event):
        menu = QMenu(self)
        menu.setObjectName("binsync_context_table_context_menu")
        
        func_addr = self.ctx if self.ctx else None
        selected_row = self.rowAt(event.pos().y())
        item = self.item(selected_row, 0)
        if item is None:
            return
        username = item.text()
        menu.addAction("Sync", lambda: self.controller.fill_function(func_addr, user=username))
        #menu.addAction("Manual Merge", lambda: manual_merge(self.controller, func_addr, user=username))

        menu.popup(self.mapToGlobal(event.pos()))


    def update_table(self, new_ctx=None):
        # only functions currently supported
        if self.ctx is None and new_ctx is None:
            return

        self.ctx = new_ctx or self.ctx
        self.items = []
        for user in self.controller.users():
            state = self.controller.client.get_state(user=user.name)

            func = state.get_function(self.ctx)

            if not func or not func.last_change:
                continue

            # changes is not currently supported
            self.items.append(
                QCTXItem(user.name, func.name, func.last_change, 0)
            )



import re
import logging
from collections import OrderedDict, defaultdict, ChainMap
from typing import Optional

import pycparser
from pycparser.c_parser import ParseError

# pycparser hack to parse type expressions
errorlog = logging.getLogger(name=__name__ + ".yacc")
errorlog.setLevel(logging.ERROR)


l = logging.getLogger(__name__)


class BSType:
    def __init__(self,
                 type_=None,
                 size=0,
                 is_primitive=True,
                 is_array=False,
                 is_ptr=False,
                 is_unknown=False
                 ):
        self.type = type_
        self._size = size

        self.is_primitive = is_primitive
        self.is_array = is_array
        self.is_ptr = is_ptr
        self.is_unknown = is_unknown

    def __str__(self):
        return f"<BSType: {self.type} {'[]' if self.is_array else ''}{'*' if self.is_ptr else ''}{'U' if self.is_unknown else ''} ({self._size})>"

    def __repr__(self):
        return self.__str__()

    @property
    def type_str(self):
        if isinstance(self.type, BSType) and self.is_array:
            return self.type.type_str + f"[{self._size}]"

        return self.type

    @property
    def base_type(self):
        if isinstance(self.type, str):
            return self
        elif isinstance(self.type, BSType):
            return self.type.base_type

        return self.type

    @property
    def size(self):
        if isinstance(self.type, BSType) and self.is_array:
            return self.type.size * self._size

        return self._size


class BSTypeParser:
    """
    Most of this code is ripped from angr's sim_type:
    https://github.com/angr/angr/blob/master/angr/sim_type.py

    It is highly simplified and drops a lot of support for real declaration parsing (like a struct dec).
    Instead, we just use it to parse types.
    """
    def __init__(self,
                 sizeof_ptr=8,
                 sizeof_long=8,
                 sizeof_double=8,
                 sizeof_int=4,
                 sizeof_float=4,
                 sizeof_short=2,
                 sizeof_char=1,
                 sizeof_bool=1,
                 extra_types=None
                 ):

        # sizes
        self.sizeof_ptr = sizeof_ptr
        self.sizeof_long = sizeof_long
        self.sizeof_double = sizeof_double
        self.sizeof_int = sizeof_int
        self.sizeof_float = sizeof_float
        self.sizeof_short = sizeof_short
        self.sizeof_char = sizeof_char
        self.sizeof_bool = sizeof_bool

        # hack in type parsing
        self._type_parser_singleton = pycparser.CParser()
        self._type_parser_singleton.cparser = pycparser.ply.yacc.yacc(
            module=self._type_parser_singleton,
            start='parameter_declaration',
            debug=False,
            optimize=False,
            errorlog=errorlog
        )
        self.ALL_TYPES = {}
        self.BASIC_TYPES = {}
        self.STDINT_TYPES = {}
        self.extra_types = extra_types or {}
        self._init_all_types()

    def _init_all_types(self):
        self.BASIC_TYPES = {
            "char": BSType(type_="char", size=self.sizeof_char),
            "signed char": BSType(type_="signed char", size=self.sizeof_char),
            "unsigned char": BSType(type_="unsigned char", size=self.sizeof_char),
            "short": BSType(type_="short", size=self.sizeof_short),
            "signed short": BSType(type_="signed short", size=self.sizeof_short),
            "unsigned short": BSType(type_="unsigned short", size=self.sizeof_short),
            "short int": BSType(type_="short int", size=self.sizeof_short),
            "signed short int": BSType(type_="signed short int", size=self.sizeof_short),
            "unsigned short int": BSType(type_="unsigned short int", size=self.sizeof_short),
            "int": BSType(type_="int", size=self.sizeof_int),
            "signed": BSType(type_="signed", size=self.sizeof_int),
            "unsigned": BSType(type_="unsigned", size=self.sizeof_int),
            "signed int": BSType(type_="signed int", size=self.sizeof_int),
            "unsigned int": BSType(type_="unsigned int", size=self.sizeof_int),
            "long": BSType(type_="long", size=self.sizeof_long),
            "signed long": BSType(type_="signed long", size=self.sizeof_long),
            "long signed": BSType(type_="long signed", size=self.sizeof_long),
            "unsigned long": BSType(type_="unsigned long", size=self.sizeof_long),
            "long int": BSType(type_="long int", size=self.sizeof_long),
            "signed long int": BSType(type_="signed long int", size=self.sizeof_long),
            "unsigned long int": BSType(type_="unsigned long int", size=self.sizeof_long),
            "long unsigned int": BSType(type_="long unsigned int", size=self.sizeof_long),
            "long long": BSType(type_="long long", size=self.sizeof_long),
            "signed long long": BSType(type_="signed long long", size=self.sizeof_long),
            "unsigned long long": BSType(type_="unsigned long long", size=self.sizeof_long),
            "long long int": BSType(type_="long long int", size=self.sizeof_long),
            "signed long long int": BSType(type_="signed long long int", size=self.sizeof_long),
            "unsigned long long int": BSType(type_="unsigned long long int", size=self.sizeof_long),
            "__int128": BSType(type_="__int128", size=16),
            "unsigned __int128": BSType(type_="unsigned __int128", size=16),
            "__int256": BSType(type_="__int256", size=32),
            "unsigned __int256": BSType(type_="unsigned __int256", size=32),
            "bool": BSType(type_="bool", size=self.sizeof_bool),
            "_Bool": BSType(type_="_Bool", size=self.sizeof_bool),
            "float": BSType(type_="float", size=self.sizeof_float),
            "double": BSType(type_="double", size=self.sizeof_double),
            "long double": BSType(type_="double", size=self.sizeof_double),
            "void": BSType(type_="void", size=self.sizeof_ptr),
        }
        self.ALL_TYPES.update(self.BASIC_TYPES)

        self.STDINT_TYPES = {
            "int8_t": BSType(type_="int8_t", size=1),
            "uint8_t": BSType(type_="uint8_t", size=1),
            "byte": BSType(type_="byte", size=1),
            "int16_t": BSType(type_="int16_t", size=2),
            "uint16_t": BSType(type_="uint16_t", size=2),
            "word": BSType(type_="word", size=2),
            "int32_t": BSType(type_="int32_t", size=4),
            "uint32_t": BSType(type_="uint32_t", size=4),
            "dword": BSType(type_="dword", size=4),
            "int64_t": BSType(type_="int64_t", size=8),
            "uint64_t": BSType(type_="uint64_t", size=8),
        }
        self.ALL_TYPES.update(self.STDINT_TYPES)
        self.ALL_TYPES.update(self.extra_types)

    def parse_type(self, defn, preprocess=True, predefined_types=None, arch=None):  # pylint:disable=unused-argument
        """
        Parse a simple type expression into a SimType

        >>> self.parse_type('int *')
        """
        return self.parse_type_with_name(defn, preprocess=preprocess, predefined_types=predefined_types, arch=arch)[0]

    def parse_type_with_name(self, defn, preprocess=True, predefined_types=None, arch=None):  # pylint:disable=unused-argument
        """
        Parse a simple type expression into a SimType, returning the a tuple of the type object and any associated name
        that might be found in the place a name would go in a type declaration.

        >>> self.parse_type_with_name('int *foo')
        """
        if not defn:
            return None

        if pycparser is None:
            raise ImportError("Please install pycparser in order to parse C definitions")

        if preprocess:
            defn = re.sub(r"/\*.*?\*/", r"", defn)

        failed_parse = False
        try:
            node = self._type_parser_singleton.parse(text=defn)
        except ParseError:
            failed_parse = True

        #
        # in the event of a failed type parse it may just be a custom type, so we should try again
        # with the struct specifier and see if it works out
        #
        if failed_parse:
            try:
                node = self._type_parser_singleton.parse(text="struct " + defn)
            except Exception:
                return (None, )

        if not isinstance(node, pycparser.c_ast.Typename) and \
                not isinstance(node, pycparser.c_ast.Decl):
            raise pycparser.c_parser.ParseError("Got an unexpected type out of pycparser")

        decl = node.type
        extra_types = {} if not predefined_types else dict(predefined_types)
        return self._decl_to_type(decl, extra_types=extra_types), node.name

    def _decl_to_type(self, decl, extra_types=None) -> Optional[BSType]:
        if not decl:
            return decl

        if extra_types is None: extra_types = {}

        if isinstance(decl, pycparser.c_ast.FuncDecl):
            return None

        elif isinstance(decl, pycparser.c_ast.TypeDecl):
            return self._decl_to_type(decl.type, extra_types)

        elif isinstance(decl, pycparser.c_ast.PtrDecl):
            pts_to = self._decl_to_type(decl.type, extra_types)
            if not pts_to:
                return None

            return BSType(type_=pts_to.type, size=self.sizeof_ptr, is_ptr=True, is_unknown=pts_to.is_unknown)

        elif isinstance(decl, pycparser.c_ast.ArrayDecl):
            elem_type = self._decl_to_type(decl.type, extra_types)

            if decl.dim is None:
                """
                r = SimTypeArray(elem_type)
                r._arch = arch
                return r
                """
                return BSType(type_=elem_type, is_array=True, size=0)
            try:
                size = self._parse_const(decl.dim, extra_types=extra_types)
            except ValueError as e:
                #l.warning("Got error parsing array dimension, defaulting to zero: %s", e)
                size = 0
            """
            r = SimTypeFixedSizeArray(elem_type, size)
            r._arch = arch
            """
            return BSType(type_=elem_type, is_array=True, size=size)

        elif isinstance(decl, pycparser.c_ast.Struct):
            if decl is None:
                return None

            return BSType(type_=decl.name, is_unknown=True)

        elif isinstance(decl, pycparser.c_ast.Union):
            return None

        elif isinstance(decl, pycparser.c_ast.IdentifierType):
            key = ' '.join(decl.names)
            if key in extra_types:
                return extra_types[key]
            elif key in self.ALL_TYPES:
                return self.ALL_TYPES[key]
            else:
                #raise TypeError("Unknown type '%s'" % key)
                return BSType(type_=key, is_unknown=True)

        elif isinstance(decl, pycparser.c_ast.Enum):
            # See C99 at 6.7.2.2
            return self.ALL_TYPES['int']

        raise ValueError("Unknown type!")

    def _make_scope(self, predefined_types=None):
        """
        Generate CParser scope_stack argument to parse method
        """
        all_types = ChainMap(predefined_types or {}, self.ALL_TYPES)
        scope = dict()
        for ty in all_types:
            if ty in self.BASIC_TYPES:
                continue
            if ' ' in ty:
                continue

            typ = all_types[ty]
            scope[ty] = True
        return [scope]

    def _parse_const(self, c, extra_types=None):
        if type(c) is pycparser.c_ast.Constant:
            return int(c.value, base=0)
        elif type(c) is pycparser.c_ast.BinaryOp:
            if c.op == '+':
                return self._parse_const(c.children()[0][1], extra_types=extra_types) + self._parse_const(
                    c.children()[1][1], extra_types=extra_types)
            if c.op == '-':
                return self._parse_const(c.children()[0][1], extra_types=extra_types) - self._parse_const(
                    c.children()[1][1], extra_types=extra_types)
            if c.op == '*':
                return self._parse_const(c.children()[0][1], extra_types=extra_types) * self._parse_const(
                    c.children()[1][1], extra_types=extra_types)
            if c.op == '/':
                return self._parse_const(c.children()[0][1], extra_types=extra_types) // self._parse_const(
                    c.children()[1][1], extra_types=extra_types)
            if c.op == '<<':
                return self._parse_const(c.children()[0][1], extra_types=extra_types) << self._parse_const(
                    c.children()[1][1], extra_types=extra_types)
            if c.op == '>>':
                return self._parse_const(c.children()[0][1], extra_types=extra_types) >> self._parse_const(
                    c.children()[1][1], extra_types=extra_types)
            raise ValueError('Binary op %s' % c.op)
        elif type(c) is pycparser.c_ast.UnaryOp:
            if c.op == 'sizeof':
                return self._decl_to_type(c.expr.type, extra_types=extra_types).size
            else:
                raise ValueError("Unary op %s" % c.op)
        elif type(c) is pycparser.c_ast.Cast:
            return self._parse_const(c.expr, extra_types=extra_types)
        else:
            raise ValueError(c)



import os
from binsync import __version__ as _VERSION

VERSION = _VERSION
PLUGIN_DIR = os.path.dirname(os.path.realpath(__file__))
IDA_DIR = os.path.abspath(os.path.join(PLUGIN_DIR, "..", ".."))
UI_DIR = os.path.join(PLUGIN_DIR, "ui")
# THEMES_DIR = os.path.join(PLUGIN_DIR, 'themes')



from binsync.common.ui.version import ui_version

if ui_version == "PySide2":
    from PySide2.QtCore import (
        QDir, Qt, Signal, QAbstractTableModel, QModelIndex, QSortFilterProxyModel, QPersistentModelIndex,
        QEvent, QThread
    )
    from PySide2.QtWidgets import (
        QAbstractItemView,
        QCheckBox,
        QComboBox,
        QDialog,
        QFileDialog,
        QGridLayout,
        QGroupBox,
        QHBoxLayout,
        QHeaderView,
        QLabel,
        QLineEdit,
        QMenu,
        QMessageBox,
        QPushButton,
        QStatusBar,
        QTableWidget,
        QTableWidgetItem,
        QTabWidget,
        QVBoxLayout,
        QWidget,
        QDialogButtonBox,
        QTableView,
        QAction,
        QFontDialog,
        QCheckBox,
        QMainWindow,
        QApplication
    )
    from PySide2.QtGui import (
        QFontDatabase,
        QColor,
        QKeyEvent,
        QFocusEvent,
        QIntValidator
    )
elif ui_version == "PySide6":
    from PySide6.QtCore import (
        QDir, Qt, Signal, QAbstractTableModel, QModelIndex, QSortFilterProxyModel, QPersistentModelIndex,
        QEvent, QThread
    )
    from PySide6.QtWidgets import (
        QAbstractItemView,
        QCheckBox,
        QComboBox,
        QDialog,
        QFileDialog,
        QGridLayout,
        QGroupBox,
        QHBoxLayout,
        QHeaderView,
        QLabel,
        QLineEdit,
        QMenu,
        QMessageBox,
        QPushButton,
        QStatusBar,
        QTableWidget,
        QTableWidgetItem,
        QTabWidget,
        QVBoxLayout,
        QWidget,
        QDialogButtonBox,
        QTableView,
        QFontDialog,
        QCheckBox,
        QMainWindow,
        QApplication
    )
    from PySide6.QtGui import (
        QFontDatabase,
        QColor,
        QKeyEvent,
        QFocusEvent,
        QIntValidator,
        QAction
    )
else:
    from PyQt5.QtCore import (
        QDir, Qt, QAbstractTableModel, QModelIndex, QSortFilterProxyModel, QPersistentModelIndex,
        QEvent, QThread
    )
    from PyQt5.QtCore import pyqtSignal as Signal
    from PyQt5.QtWidgets import (
        QAbstractItemView,
        QCheckBox,
        QComboBox,
        QDialog,
        QFileDialog,
        QGridLayout,
        QGroupBox,
        QHBoxLayout,
        QHeaderView,
        QLabel,
        QLineEdit,
        QMenu,
        QMessageBox,
        QPushButton,
        QStatusBar,
        QTableWidget,
        QTableWidgetItem,
        QTabWidget,
        QVBoxLayout,
        QWidget,
        QDialogButtonBox,
        QTableView,
        QAction,
        QFontDialog,
        QCheckBox,
        QMainWindow,
        QApplication
    )
    from PyQt5.QtGui import (
        QFontDatabase,
        QColor,
        QKeyEvent,
        QFocusEvent,
        QIntValidator
    )



__version__ = "2.7.1"

#
# logging
#

import logging
logging.getLogger("binsync").addHandler(logging.NullHandler())
from binsync.loggercfg import Loggers
loggers = Loggers()
del Loggers
del logging

from binsync.data import *
from binsync import common
from binsync.common import BinSyncController, SyncControlStatus
from binsync.core.client import Client, ConnectionWarnings, BINSYNC_ROOT_BRANCH



ui_version = "PySide2"


def set_ui_version(version):
    global ui_version
    valid_version = [
        "PyQt5",
        "PySide2",
        "PySide6"
    ]

    if version in valid_version:
        ui_version = version
    else:
        raise Exception("Failed to set BinSync UI version")



import logging

from binsync.common import ArtifactLifter

from . import compat

l = logging.getLogger(name=__name__)


class IDAArtifactLifter(ArtifactLifter):
    lift_map = {
        "__int64": "long",
        "__int32": "int",
        "__int16": "short",
        "__int8": "char",
    }

    def __init__(self, controller):
        super(IDAArtifactLifter, self).__init__(controller)

    def lift_addr(self, addr: int) -> int:
        return addr

    def lift_type(self, type_str: str) -> str:
        for ida_t, bs_t in self.lift_map.items():
            type_str = type_str.replace(ida_t, bs_t)

        return type_str

    def lift_stack_offset(self, offset: int, func_addr: int) -> int:
        return offset

    def lower_addr(self, addr: int) -> int:
        return addr

    def lower_type(self, type_str: str) -> str:
        return type_str

    def lower_stack_offset(self, offset: int, func_addr: int) -> int:
        return offset #compat.ida_to_angr_stack_offset(func_addr, offset)



import logging
import threading
import time
from functools import wraps
from typing import Dict, Iterable, List, Optional, Union

import binsync.data
from binsync.common.artifact_lifter import ArtifactLifter
from binsync.core.client import Client, SchedSpeed, Scheduler, Job
from binsync.data.type_parser import BSTypeParser, BSType
from binsync.data import (
    State, User, Artifact,
    Function, FunctionHeader, FunctionArgument, StackVariable,
    Comment, GlobalVariable, Patch,
    Enum, Struct
)

_l = logging.getLogger(name=__name__)


#
# State Checking Decorators
#

def init_checker(f):
    @wraps(f)
    def _init_check(self, *args, **kwargs):
        if not self.check_client():
            raise RuntimeError("Please connect to a repo first.")
        return f(self, *args, **kwargs)

    return _init_check

def fill_event(f):
    @wraps(f)
    def _fill_event(self: "BinSyncController", *args, **kwargs):
        return self.fill_event_handler(f, *args, **kwargs)

    return _fill_event

#
# Description Constants
#

# https://stackoverflow.com/questions/10926328
BUSY_LOOP_COOLDOWN = 0.5
GET_MANY = True
FILL_MANY = True

class SyncControlStatus:
    CONNECTED = 0
    CONNECTED_NO_REMOTE = 1
    DISCONNECTED = 2


class MergeLevel:
    OVERWRITE = 0
    NON_CONFLICTING = 1
    MERGE = 2


class FakeSyncLock:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


#
#   Controller
#

class BinSyncController:

    ARTIFACT_SET_MAP = {
        Function: State.set_function,
        FunctionHeader: State.set_function_header,
        StackVariable: State.set_stack_variable,
        Comment: State.set_comment,
        GlobalVariable: State.set_global_var,
        Struct: State.set_struct,
        Enum: State.set_enum
    }

    ARTIFACT_GET_MAP = {
        Function: State.get_function,
        (Function, GET_MANY): State.get_functions,
        FunctionHeader: State.get_function_header,
        (FunctionHeader, GET_MANY): State.get_function_headers,
        StackVariable: State.get_stack_variable,
        (StackVariable, GET_MANY): State.get_stack_variables,
        Comment: State.get_comment,
        (Comment, GET_MANY): State.get_func_comments,
        GlobalVariable: State.get_global_var,
        (GlobalVariable, GET_MANY): State.get_global_vars,
        Struct: State.get_struct,
        (Struct, GET_MANY): State.get_structs,
        Enum: State.get_enum,
        (Enum, GET_MANY): State.get_enums,
    }

    """
    The BinSync Controller is the main interface for syncing with the BinSync Client which preforms git tasks
    such as pull and push. In the Controller higher-level tasks are done such as updating UI with changes
    and preforming syncs and pushes on data users need/change.

    All class properties that have a "= None" means they must be set during runtime by an outside process.
    The client will be set on connection. The ctx_change_callback will be set by an outside UI

    """
    def __init__(self, artifact_lifter=None, headless=False, reload_time=10):
        self.headless = headless
        self.reload_time = reload_time
        self.artifact_lifer: ArtifactLifter = artifact_lifter

        # client created on connection
        self.client = None  # type: Optional[Client]

        # ui callback created on UI init
        self.ui_callback = None  # func()
        self.ctx_change_callback = None  # func()
        self._last_reload = None
        self.last_ctx = None
        self._init_ui_components()

        # settings
        self.merge_level: int = MergeLevel.NON_CONFLICTING

        # command locks
        self.job_scheduler = Scheduler()
        self.sync_lock = threading.Lock()

        # create a pulling thread, but start on connection
        self.updater_thread = threading.Thread(target=self.updater_routine)
        self._run_updater_threads = False

        # TODO: make the initialization of this with types of decompiler
        self.type_parser = BSTypeParser()

    #
    #   Multithreading updaters, locks, and evaluators
    #

    def _init_ui_components(self):
        if self.headless:
            return

        # after this point you can import anything from UI and it is safe!
        from binsync.common.ui.qt_objects import (
            QThread
        )

    def schedule_job(self, cmd_func, *args, blocking=False, **kwargs):
        if blocking:
            return self.job_scheduler.schedule_and_wait_job(
                Job(cmd_func, *args, **kwargs)
            )

        self.job_scheduler.schedule_job(
            Job(cmd_func, *args, **kwargs)
        )
        return None

    def updater_routine(self):
        while self._run_updater_threads:
            time.sleep(BUSY_LOOP_COOLDOWN)

            # validate a client is connected to this controller (may not have remote )
            if not self.check_client():
                continue

            # do git pull/push operations if a remote exist for the client
            if self.client.last_pull_attempt_ts is None:
                self.client.update(commit_msg="User created")

            # update every reload_time
            elif time.time() - self.client.last_pull_attempt_ts > self.reload_time:
                self.client.update()

            if not self.headless:
                # update context knowledge every loop iteration
                if self.ctx_change_callback:
                    self._check_and_notify_ctx()

                # update the control panel with new info every BINSYNC_RELOAD_TIME seconds
                if self._last_reload is None or \
                        time.time() - self._last_reload > self.reload_time:
                    self._last_reload = time.time()
                    self._update_ui()

    def _update_ui(self):
        if not self.ui_callback:
            return

        self.ui_callback()

    def _check_and_notify_ctx(self):
        active_ctx = self.active_context()
        if active_ctx is None or self.last_ctx == active_ctx:
            return

        self.last_ctx = active_ctx
        self.ctx_change_callback()

    def start_worker_routines(self):
        self._run_updater_threads = True
        self.updater_thread.setDaemon(True)
        self.updater_thread.start()

        self.job_scheduler.start_worker_thread()

    def stop_worker_routines(self):
        self._run_updater_threads = False
        self.job_scheduler.stop_worker_thread()

    #
    # Client Interaction Functions
    #

    def connect(self, user, path, init_repo=False, remote_url=None):
        binary_hash = self.binary_hash()
        self.client = Client(
            user, path, binary_hash, init_repo=init_repo, remote_url=remote_url
        )

        self.start_worker_routines()
        return self.client.connection_warnings

    def check_client(self):
        return self.client is not None

    def status(self):
        if self.check_client():
            if self.client.has_remote and self.client.active_remote:
                return SyncControlStatus.CONNECTED
            return SyncControlStatus.CONNECTED_NO_REMOTE
        return SyncControlStatus.DISCONNECTED

    def status_string(self):
        stat = self.status()
        if stat == SyncControlStatus.CONNECTED:
            return f"<font color=#1eba06>{self.client.master_user}</font>"
        elif stat == SyncControlStatus.CONNECTED_NO_REMOTE:
            return f"<font color=#e7b416>{self.client.master_user}</font>"
        else:
            return "<font color=#cc3232>Disconnected</font>"

    def toggle_headless(self):
        self.headless = not self.headless

    @init_checker
    def users(self, priority=None) -> Iterable[User]:
        return self.client.users(priority=priority)

    def usernames(self, priority=None) -> Iterable[str]:
        for user in self.users(priority=priority):
            yield user.name

    #
    # Override Mandatory API:
    # These functions create a public API for things that hold a reference to the Controller from either another
    # thread or object. This is most useful for use in the UI, which can use this API to make general requests from
    # the decompiler regardless of internal decompiler API.
    #

    def binary_hash(self) -> str:
        """
        Returns a hex string of the currently loaded binary in the decompiler. For most cases,
        this will simply be a md5hash of the binary.

        @rtype: hex string
        """
        raise NotImplementedError

    def active_context(self) -> binsync.data.Function:
        """
        Returns an binsync Function. Currently only functions are supported as current contexts.
        This function will be called very frequently, so its important that its implementation is fast
        and can be done many times in the decompiler.
        """
        raise NotImplementedError

    def binary_path(self) -> Optional[str]:
        """
        Returns a string that is the path of the currently loaded binary. If there is no binary loaded
        then None should be returned.

        @rtype: path-like string (/path/to/binary)
        """
        raise NotImplementedError

    def get_func_size(self, func_addr) -> int:
        """
        Returns the size of a function

        @param func_addr:
        @return:
        """
        raise NotImplementedError

    def goto_address(self, func_addr) -> None:
        """
        Relocates decompiler display to provided address

        @param func_addr:
        @return:
        """
        raise NotImplementedError

    @property
    def decompiler_available(self) -> bool:
        """
        @return: True if decompiler is available for decompilation, False if otherwise
        """
        return True

    #
    # Optional Artifact API:
    # A series of functions that allow public access to live artifacts in the decompiler. As an example,
    # `function(addr)` will return the current Function at addr that the user would be seeing. This is useful
    # for having a common interface of reading data from other decompilers.
    #

    def functions(self) -> Dict[int, Function]:
        """
        Returns a dict of binsync.Functions that contain the addr, name, and size of each function in the decompiler.
        Note: this does not contain the live data of the Artifact, only the minimum knowledge to that the Artifact
        exists. To get live data, use the singleton function of the same name.

        @return:
        """
        return {}

    def function(self, addr) -> Optional[Function]:
        return None

    def global_vars(self) -> Dict[int, GlobalVariable]:
        """
        Returns a dict of binsync.GlobalVariable that contain the addr and size of each global var.
        Note: this does not contain the live data of the Artifact, only the minimum knowledge to that the Artifact
        exists. To get live data, use the singleton function of the same name.

        @return:
        """
        return {}

    def global_var(self, addr) -> Optional[GlobalVariable]:
        return None

    def structs(self) -> Dict[str, Struct]:
        """
        Returns a dict of binsync.Structs that contain the name and size of each struct in the decompiler.
        Note: this does not contain the live data of the Artifact, only the minimum knowledge to that the Artifact
        exists. To get live data, use the singleton function of the same name.

        @return:
        """
        return {}

    def struct(self, name) -> Optional[Struct]:
        return None

    def enums(self) -> Dict[str, Enum]:
        """
        Returns a dict of binsync.Enum that contain the name of the enums in the decompiler.
        Note: this does not contain the live data of the Artifact, only the minimum knowledge to that the Artifact
        exists. To get live data, use the singleton function of the same name.

        @return:
        """
        return {}

    def enum(self, name) -> Optional[Enum]:
        return None

    def patches(self) -> Dict[int, Patch]:
        """
        Returns a dict of binsync.Patch that contain the addr of each Patch and the bytes.
        Note: this does not contain the live data of the Artifact, only the minimum knowledge to that the Artifact
        exists. To get live data, use the singleton function of the same name.

        @return:
        """
        return {}

    def patch(self, addr) -> Optional[Patch]:
        return None

    def global_artifacts(self):
        """
        Returns a light version of all artifacts that are global (non function associated):
        - structs, gvars, enums

        @return:
        """
        g_artifacts = {}
        for f in [self.structs, self.global_vars, self.enums]:
            g_artifacts.update(f())

        return g_artifacts

    def global_artifact(self, lookup_item: Union[str, int]):
        """
        Returns a live binsync.data version of the Artifact located at the lookup_item location, which can
        lookup any artifact supported in `global_artifacts`

        @param lookup_item:
        @return:
        """

        if isinstance(lookup_item, int):
            return self.global_var(lookup_item)
        elif isinstance(lookup_item, str):
            artifact = self.struct(lookup_item)
            if artifact:
                return artifact

            artifact = self.enum(lookup_item)
            return artifact

        return None

    #
    # Client API & Shortcuts
    #

    def lift_artifact(self, artifact: Artifact) -> Artifact:
        return self.artifact_lifer.lift(artifact)

    def lower_artifact(self, artifact: Artifact) -> Artifact:
        return self.artifact_lifer.lower(artifact)

    @init_checker
    def get_state(self, user=None, version=None, priority=None, no_cache=False) -> State:
        return self.client.get_state(user=user, version=version, priority=priority, no_cache=no_cache)

    @init_checker
    def pull_artifact(self, type_: Artifact, *identifiers, many=False, user=None, state=None) -> Optional[Artifact]:
        try:
            get_artifact_func = self.ARTIFACT_GET_MAP[type_] if not many else self.ARTIFACT_GET_MAP[(type_, GET_MANY)]
        except KeyError:
            _l.info(f"Attempting to pull an unsupported Artifact of type {type_} with {identifiers}")
            return None

        # assure a state exists
        if not state:
            state = self.get_state(user=user)

        try:
            artifact = get_artifact_func(state, *identifiers)
        except Exception:
            _l.warning(f"Failed to pull an supported Artifact of type {type_} with {identifiers}")
            return None

        if not artifact:
            return artifact

        return self.lower_artifact(artifact)

    @init_checker
    def push_artifact(self, artifact: Artifact, user=None, state=None, commit_msg=None, set_last_change=True, **kwargs) -> bool:
        """
        Every pusher artifact does three things
        1. Get the state setter function based on the class of the Obj
        2. Get the commit msg of the obj based on the class
        3. Lift the obj based on the Controller lifters

        @param artifact:
        @param user:
        @param state:
        @return:
        """
        _l.debug(f"Running push now for {artifact}")
        if not artifact:
            return False

        try:
            set_artifact_func = self.ARTIFACT_SET_MAP[artifact.__class__]
        except KeyError:
            _l.info(f"Attempting to push an unsupported Artifact of type {artifact}")
            return False

        # assure state exists
        if not state:
            state = self.get_state(user=user)

        # assure function existence for artifacts requiring a function
        if isinstance(artifact, (FunctionHeader, StackVariable, Comment)):
            func_addr = artifact.func_addr if hasattr(artifact, "func_addr") else artifact.addr
            if func_addr and not state.get_function(func_addr):
                self.push_artifact(Function(func_addr, self.get_func_size(func_addr)), state=state, set_last_change=set_last_change)

        # lift artifact into standard BinSync format
        artifact = self.lift_artifact(artifact)

        # set the artifact in the target state, likely master
        _l.debug(f"Setting an artifact now into {state} as {artifact}")
        was_set = set_artifact_func(state, artifact, set_last_change=set_last_change)

        # TODO: make was_set reliable
        _l.debug(f"{state} committing now with {commit_msg or artifact.commit_msg}")
        self.client.commit_state(state, msg=commit_msg or artifact.commit_msg)

        return was_set

    #
    # Fillers:
    # A filler function is generally responsible for pulling down data from a specific user state
    # and reflecting those changes in decompiler view (like the text on the screen). Normally, these changes
    # will also be accompanied by a Git commit to the master users state to save the changes from pull and
    # fill into their BS database. In special cases, a filler may only update the decompiler UI but not directly
    # cause a save of the BS state.
    #

    def fill_event_handler(self, filler_func, *identifiers,
                           artifact=None, user=None, state=None, master_state=None, merge_level=None, blocking=False,
                           commit_msg=None,
                           **kwargs
                           ):
        """
        fill_event_handler is the function called before every `fill_<artifact>` function. This handler is responsible
        for assuring a variety of things exist for subsequent call to `fill_<artifact>`.

        @param filler_func:
        @param identifiers:
        @param artifact:
        @param user:
        @param state:
        @param master_state:
        @param merge_level:
        @param blocking:
        @param kwargs:
        @return:
        """

        ARTIFACT_FILL_MAP = {
            self.fill_function.__name__: Function,
            self.fill_function_header.__name__: FunctionHeader,
            self.fill_stack_variable.__name__: StackVariable,
            self.fill_comment.__name__: Comment,
            self.fill_global_var.__name__: GlobalVariable,
            self.fill_struct.__name__: Struct,
            self.fill_enum.__name__: Enum
        }

        state = state if state is not None else self.get_state(user=user, priority=SchedSpeed.FAST)
        master_state = master_state if master_state is not None else self.get_state(priority=SchedSpeed.FAST)
        artifact_type = ARTIFACT_FILL_MAP.get(filler_func.__name__, None)
        if not artifact_type:
            _l.warning(f"Attempting to Fill an unknown type! Stopping Fill...")
            return None

        art_getter = self.ARTIFACT_GET_MAP.get(artifact_type)
        master_artifact = artifact if artifact else self.lower_artifact(art_getter(master_state, *identifiers))
        merged_artifact = self.merge_artifacts(
            master_artifact,  self.lower_artifact(art_getter(state, *identifiers)),
            merge_level=merge_level, master_state=master_state
        )

        lock = self.sync_lock if not self.sync_lock.locked() else FakeSyncLock()
        with lock:
            fill_changes = filler_func(
                self,
                *identifiers,
                artifact=merged_artifact, user=user, state=state, master_state=master_state, merge_level=merge_level,
                **kwargs
            )

        _l.info(
            f"Successfully synced new changes from {state.user} for {merged_artifact}" if fill_changes
            else f"No new changes or failed to sync from {state.user} for {merged_artifact}"
        )

        if blocking:
            self.push_artifact(merged_artifact, state=master_state, set_last_change=False, commit_msg=commit_msg)
        else:
            self.schedule_job(
                self.push_artifact,
                merged_artifact,
                state=master_state,
                set_last_change=False,
                commit_msg=commit_msg
            )

        return fill_changes


    @init_checker
    @fill_event
    def fill_struct(self, struct_name, header=True, members=True, artifact=None, **kwargs):
        """

        @param struct_name:
        @param user:
        @param state:
        @param header:
        @param members:
        @return:
        """
        _l.debug(f"Fill Struct is not implemented in your decompiler.")
        return False

    @init_checker
    @fill_event
    def fill_global_var(self, var_addr, user=None, artifact=None, **kwargs):
        """
        Grab a global variable for a specified address and fill it locally

        @param var_addr:
        @param user:
        @param state:
        @return:
        """
        _l.debug(f"Fill Global Var is not implemented in your decompiler.")
        return False


    @init_checker
    @fill_event
    def fill_enum(self, enum_name, user=None, artifact=None, **kwargs):
        """
        Grab an enum and fill it locally

        @param enum_name:
        @param user:
        @param state:
        @return:
        """
        _l.debug(f"Fill Enum is not implemented in your decompiler.")
        return False

    @fill_event
    def fill_stack_variable(self, func_addr, offset, user=None, artifact=None, **kwargs):
        _l.debug(f"Fill Stack Var is not implemented in your decompiler.")
        return False

    @fill_event
    def fill_function_header(self, func_addr, user=None, artifact=None, **kwargs):
        _l.debug(f"Fill Function Header is not implemented in your decompiler.")
        return False

    @fill_event
    def fill_comment(self, addr, user=None, artifact=None, **kwargs):
        _l.debug(f"Fill Comments is not implemented in your decompiler.")
        return False

    @init_checker
    @fill_event
    def fill_function(self, func_addr, user=None, artifact=None, **kwargs):
        """
        Grab all relevant information from the specified user and fill the @func_addr.
        """
        dec_func: Function = self.function(func_addr)
        if not dec_func:
            _l.warning(f"The function at {hex(func_addr)} does not exist in your decompiler.")
            dec_func = Function(None, None)

        master_func: Function = artifact
        changes = False

        # function header
        if master_func.header and master_func.header != dec_func.header:
            # type is user made (a struct)
            changes |= self.import_user_defined_type(master_func.header.ret_type, **kwargs)
            changes |= self.fill_function_header(func_addr, artifact=master_func.header, **kwargs)

        # stack vars
        if master_func.stack_vars and master_func.stack_vars != dec_func.stack_vars:
            for offset, sv in master_func.stack_vars.items():
                dec_sv = dec_func.stack_vars.get(offset, None)
                if not dec_sv:
                    _l.warning(f"Decompiler stack var not found at {offset}")

                if dec_sv == sv:
                    continue

                corrected_off = dec_sv.stack_offset if dec_sv and dec_sv.stack_offset else sv.stack_offset
                changes |= self.import_user_defined_type(sv.type, **kwargs)
                changes |= self.fill_stack_variable(func_addr, corrected_off, artifact=sv, **kwargs)

        # comments
        for addr, cmt in kwargs['state'].get_func_comments(func_addr).items():
            changes |= self.fill_comment(addr, artifact=cmt, **kwargs)

        return changes

    @init_checker
    def fill_functions(self, user=None, **kwargs):
        change = False
        master_state, state = self.get_master_and_user_state(user=user, **kwargs)
        for addr, func in state.functions.items():
            change |= self.fill_function(addr, state=state, master_state=master_state)

        return change

    @init_checker
    def fill_structs(self, user=None, **kwargs):
        """
        Grab all the structs from a specified user, then fill them locally

        @param user:
        @param state:
        @return:
        """
        changes = False
        # only do struct headers for circular references
        master_state, state = self.get_master_and_user_state(user=user, **kwargs)
        for name, struct in state.structs.items():
            changes |= self.fill_struct(name, user=user, state=state, master_state=master_state, members=False)

        for name, struct in state.structs.items():
            changes |= self.fill_struct(name, user=user, state=state, master_state=master_state, header=False)

        return changes

    @init_checker
    def fill_enums(self, user=None, **kwargs):
        """
        Grab all enums and fill it locally

        @param user:
        @param state:
        @return:
        """
        changes = False
        master_state, state = self.get_master_and_user_state(user=user, **kwargs)
        for name, enum in state.enums.items():
            changes |= self.fill_enum(name, user=user, state=state, master_state=master_state)

        return changes

    @init_checker
    def fill_global_vars(self, user=None, **kwargs):
        changes = False
        master_state, state = self.get_master_and_user_state(user=user, **kwargs)
        for off, gvar in state.global_vars.items():
            changes |= self.fill_global_var(off, user=user, state=state, master_state=master_state)

        return changes

    @init_checker
    def fill_all(self, user=None, **kwargs):
        """
        Connected to the Sync All action:
        syncs in all the data from the targeted user

        @param user:
        @param state:
        @return:
        """
        _l.info(f"Filling all data from user {user}...")

        master_state, state = self.get_master_and_user_state(user=user, **kwargs)
        fillers = [
            self.fill_structs, self.fill_enums, self.fill_global_vars, self.fill_functions
        ]

        changes = False
        for filler in fillers:
            changes |= filler(user=user, state=state, master_state=master_state)

        return changes

    @init_checker
    def magic_fill(self, preference_user=None, target_artifacts=None):
        """
        Traverses all the data in the BinSync repo, starting with an optional preference user,
        and sequentially merges that data together in a non-conflicting way. This also means that the prefrence
        user makes up the majority of the initial data you sync in.

        This process supports: functions (header, stack vars), structs, and global vars
        TODO:
        - support for enums
        - refactor fill_function to stop attempting to set master state after we do
        -

        @param preference_user:
        @param target_artifacts:
        @return:
        """
        _l.info(f"Staring a Magic Sync with a preference for {preference_user}")

        if self.merge_level == MergeLevel.OVERWRITE:
            _l.warning(f"Using Magic Sync with OVERWRITE is not supported, switching to NON-CONFLICTING")

        # re-order users for the prefered user to be at the front of the queue (if they exist)
        all_users = list(self.usernames(priority=SchedSpeed.FAST))
        preference_user = preference_user if preference_user else self.client.master_user
        master_state = self.client.get_state(user=self.client.master_user, priority=SchedSpeed.FAST)
        users_state_map = {
            user: self.get_state(user=user, priority=SchedSpeed.FAST)
            for user in all_users
        }
        all_users.remove(preference_user)

        target_artifacts = target_artifacts or {
            Struct: self.fill_struct,
            Comment: lambda *x, **y: None,
            Function: self.fill_function,
            GlobalVariable: self.fill_global_var,
            Enum: self.fill_enum
        }

        for artifact_type, filler_func in target_artifacts.items():
            _l.info(f"Magic Syncing artifacts of type {artifact_type.__name__} now...")
            pref_state = users_state_map[preference_user]
            for identifier in self.changed_artifacts_of_type(artifact_type, users=all_users + [preference_user], states=users_state_map):
                pref_art = self.pull_artifact(artifact_type, identifier, state=pref_state)
                for user in all_users:
                    user_state = users_state_map[user]
                    user_art = self.pull_artifact(artifact_type, identifier, state=user_state)

                    if not user_art:
                        continue

                    if not pref_art:
                        pref_art = user_art.copy()

                    pref_art = pref_art.nonconflict_merge(user_art)
                    pref_art.last_change = None

                try:
                    filler_func(
                        identifier, artifact=pref_art, state=master_state,  commit_msg=f"Magic Synced {pref_art}",
                        merge_level=MergeLevel.NON_CONFLICTING
                    )
                except Exception as e:
                    _l.info(f"Banishing exception: {e}")

        _l.info(f"Magic Syncing Completed!")

    #
    # Force Push
    #

    @init_checker
    def force_push_function(self, addr: int) -> bool:
        """
        Collects the function currently stored in the decompiler, not the BS State, and commits it to
        the master users BS Database.

        TODO: push the comments and custom types that are associated with each stack var
        TODO: refactor to use internal push_function for correct commit message

        @param addr:
        @return: Success of committing the Function
        """
        func = self.function(addr)
        if not func:
            _l.info(f"Pushing function {hex(addr)} Failed")
            return False

        master_state: State = self.client.get_state(priority=SchedSpeed.FAST)
        pushed = self.push_artifact(func, state=master_state, commit_msg=f"Forced pushed function {func}")
        return pushed


    @init_checker
    def force_push_global_artifact(self, lookup_item):
        """
        Collects the global artifact (struct, gvar, enum) currently stored in the decompiler, not the BS State,
        and commits it to the master users BS Database.

        @param lookup_item:
        @return: Success of committing the Artifact
        """
        global_art = self.global_artifact(lookup_item)
        
        if not global_art:
            _l.info(f"Pushing global artifact {lookup_item if isinstance(lookup_item, str) else hex(lookup_item)} Failed")
            return False

        master_state: State = self.client.get_state(priority=SchedSpeed.FAST)
        global_art = self.artifact_lifer.lift(global_art)
        pushed = self.push_artifact(global_art, state=master_state, commit_msg=f"Force pushed {global_art}")
        return pushed

    #
    # Utils
    #

    def merge_artifacts(self, art1: Artifact, art2: Artifact, merge_level=None, **kwargs):
        if merge_level is None:
            merge_level = self.merge_level

        if art2 is None:
            return art1.copy()

        if merge_level == MergeLevel.OVERWRITE or (not art1) or (art1 == art2):
            return art2.copy() if art2 else None

        if merge_level == MergeLevel.NON_CONFLICTING:
            merge_art = art1.nonconflict_merge(art2, **kwargs)

        elif merge_level == MergeLevel.MERGE:
            _l.warning("Manual Merging is not currently supported, using non-conflict syncing...")
            merge_art = art1.nonconflict_merge(art2, **kwargs)

        else:
            raise Exception("Your BinSync Client has an unsupported Sync Level activated")

        return merge_art

    def changed_artifacts_of_type(self, type_: Artifact, users=[], states={}):
        prop_map = {
            Function: "functions",
            Comment: "comments",
            GlobalVariable: "global_vars",
            Struct: "structs",
            Enum: "enums"
        }

        try:
            prop_name = prop_map[type_]
        except KeyError:
            _l.warning(f"Attempted to get changed artifacts of type {type_} which is unsupported")
            return set()

        known_arts = set()
        for username in users:
            state = states[username]
            artifact_dict: Dict = getattr(state, prop_name)
            for identifier in artifact_dict:
                known_arts.add(identifier)

        return known_arts

    def type_is_user_defined(self, type_str, state=None):
        if not type_str:
            return None

        type_: BSType = self.type_parser.parse_type(type_str)
        if not type_:
            # it was not parseable
            return None

        # type is known and parseable
        if not type_.is_unknown:
            return None

        base_type_str = type_.base_type.type_str
        return base_type_str if base_type_str in state.structs.keys() else None

    def import_user_defined_type(self, type_str, **kwargs):
        state = kwargs.pop('state')
        master_state = kwargs['master_state']
        base_type_str = self.type_is_user_defined(type_str, state=state)
        if not base_type_str:
            return False

        struct: Struct = state.get_struct(base_type_str)
        if not struct:
            return False

        nested_undefined_structs = False
        for off, memb in struct.struct_members.items():
            user_type = self.type_is_user_defined(memb.type, state=state)
            if user_type and user_type not in master_state.structs.keys():
                # should we ever happen to have a struct with a nested type that is
                # also a struct that we don't have in our master_state, then we give up
                # and attempt to fill all structs to resolve type issues
                nested_undefined_structs = True
                _l.info(f"Nested undefined structs detected, pulling all structs from {state.user}")
                break

        changes = self.fill_struct(base_type_str, state=state, **kwargs) if not nested_undefined_structs \
            else self.fill_structs(state=state, **kwargs)
        return changes

    def get_master_and_user_state(self, user=None, **kwargs):
        state = kwargs.get("state", None) \
            or self.get_state(user=user, priority=SchedSpeed.FAST)

        master_state = kwargs.get("master_state", None) \
            or self.get_state(priority=SchedSpeed.FAST)

        return master_state, state



import argparse
import sys
import logging
from pathlib import Path
import importlib
import pkg_resources

import binsync

l = logging.getLogger(__name__)


def execute_plugin_entry(plugin_name):
    plugins_path = Path(
        pkg_resources.resource_filename("binsync", f"plugins")
    )
    if not plugins_path.exists():
        l.error("Known plugins path does not exist, which means BinSync did not install correctly!")
        return False

    sys.path.insert(1, str(plugins_path))
    plugin = importlib.import_module(f"{plugin_name}_binsync")
    l.debug(f"Executing {plugin_name} plugin...")
    return plugin.start()


def main():
    parser = argparse.ArgumentParser(description="BinSync command line util")
    parser.add_argument(
        "-p", "--run-plugin", help="Execute BinSync decompiler plugin by command line, results may vary!"
    )

    args = parser.parse_args()
    plugin_to_run = args.run_plugin

    if plugin_to_run:
        return execute_plugin_entry(plugin_to_run)


if __name__ == "__main__":
    main()



import logging
import time

import binsync.data
from binsync.common.ui.force_push.force_push import ForcePushUI
from binsync.common.ui.panel_tabs.activity_table import QActivityTable
from binsync.common.ui.panel_tabs.ctx_table import QCTXTable
from binsync.common.ui.panel_tabs.functions_table import QFunctionTable
from binsync.common.ui.panel_tabs.globals_table import QGlobalsTable
from binsync.common.ui.panel_tabs.util_panel import QUtilPanel
from binsync.common.ui.qt_objects import (
    QLabel,
    QMenu,
    QPushButton,
    QStatusBar,
    QTabWidget,
    QVBoxLayout,
    QWidget,
    Signal
)

l = logging.getLogger(__name__)

class QContextStatusBar(QStatusBar):
    def __init__(self, controller, parent = None):
        super(QContextStatusBar, self).__init__(parent)
        self.controller = controller

    def contextMenuEvent(self, event):
        menu = QMenu(self)
        menu.setObjectName("binsync_context_context_menu")
        if self.controller.last_ctx:
            ctx_name = self.controller.last_ctx.name
            ctx_addr = self.controller.last_ctx.addr
            menu.addAction(f"Force Push {ctx_name}@{hex(ctx_addr)}", lambda : self.controller.force_push_function(ctx_addr))
            menu.popup(self.mapToGlobal(event.pos()))

class ControlPanel(QWidget):
    update_ready = Signal()
    ctx_change = Signal()

    def __init__(self, controller, parent=None):
        super(ControlPanel, self).__init__(parent)
        self.controller = controller

        self.tables = {}
        self._init_widgets()

        # register controller callback
        self.update_ready.connect(self.reload)
        self.controller.ui_callback = self.update_callback

        self.ctx_change.connect(self._reload_ctx)
        self.controller.ctx_change_callback = self.ctx_callback

    def update_callback(self):
        """
        This function will be called in another thread, so the work
        done here is guaranteed to be thread safe.

        @return:
        """
        self._update_table_data()
        self.update_ready.emit()

    def ctx_callback(self):
        if isinstance(self.controller.last_ctx, binsync.data.Function):
            self._ctx_table.update_table(new_ctx=self.controller.last_ctx.addr)

        self.ctx_change.emit()

    def reload(self):
        # check if connected
        if self.controller and self.controller.check_client():
            self._reload_tables()

        # update status
        status = self.controller.status_string() if self.controller else "Disconnected"
        self._status_label.setText(status)

    def closeEvent(self, event):
        if self.controller is not None:
            self.controller.client_init_callback = None

    def _init_widgets(self):
        # status bar
        self._status_label = QLabel(self)
        self._status_label.setText(self.controller.status_string())
        self._status_bar = QContextStatusBar(self.controller, self)
        self._status_bar.addPermanentWidget(self._status_label)

        # control box
        control_layout = QVBoxLayout()

        # tabs for panel_tabs
        self.tabView = QTabWidget()

        # add panel_tabs to tabs
        self._ctx_table = QCTXTable(self.controller)
        self._func_table = QFunctionTable(self.controller)
        self._global_table = QGlobalsTable(self.controller)
        self._activity_table = QActivityTable(self.controller)
        self._utilities_panel = QUtilPanel(self.controller)

        self.tabView.addTab(self._ctx_table, "Context")
        self.tabView.addTab(self._func_table, "Functions")
        self.tabView.addTab(self._global_table, "Globals")
        self.tabView.addTab(self._activity_table, "Activity")
        self.tabView.addTab(self._utilities_panel, "Utilities")

        self.tables.update({
            "functions": self._func_table,
            "globals": self._global_table,
            "activity": self._activity_table
        })

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.tabView)
        main_layout.addWidget(self._status_bar)
        main_layout.setSpacing(0)
        main_layout.setContentsMargins(0,0,0,0)


        self.setLayout(main_layout)

    def _reload_ctx(self):
        ctx_name = self.controller.last_ctx.name or ""
        ctx_name = ctx_name[:12] + "..." if len(ctx_name) > 12 else ctx_name
        self._status_bar.showMessage(f"{ctx_name}@{hex(self.controller.last_ctx.addr)}")
        self._ctx_table.reload()

    def _reload_tables(self):
        for _, table in self.tables.items():
            table.reload()
        self._ctx_table.reload()

    def _update_table_data(self):
        for _, table in self.tables.items():
            table.update_table()

        self._ctx_table.update_table()



import sys
import os
import subprocess
from urllib.request import urlretrieve

# How to run:
# import urllib.request, sys, os; sys.path.insert(1, os.getcwd()); urllib.request.urlretrieve("https://raw.githubusercontent.com/angr/binsync/one_liner_ida/plugins/ida_binsync/ida_binsync/oneliner.py", "oneliner.py"); from oneliner import install; install()


class PlatformType:
    MACOS = "macos"
    LINUX = "linux"
    WINDOWS = "windows"


def find_platform():
    platform = PlatformType.WINDOWS

    if sys.platform == "linux" or sys.platform == "linux2":
        platform = PlatformType.LINUX
    elif sys.platform == "darwin":
        platform = PlatformType.MACOS

    return platform


def plugin_install_binsync(plugins_path):
    github_url_base = "https://raw.githubusercontent.com/angr/binsync/master/plugins/ida_binsync/"
    ida_binsync_folder = os.path.join(plugins_path, "ida_binsync")

    # install entry point of binsync
    urlretrieve(github_url_base+"ida_binsync.py", os.path.join(plugins_path, "ida_binsync.py"))

    # install ida_binsync/*
    github_url_base += "ida_binsync/"
    files_to_download = ["__init__.py", "compat.py", "controller.py", "hooks.py", "plugin.py", "artifact_lifter.py"]
    try:
        os.mkdir(ida_binsync_folder)
    except FileExistsError:
        pass
    for f in files_to_download:
        urlretrieve(github_url_base+f, os.path.join(ida_binsync_folder, f))


def pip_install_binsync(python_path):
    subprocess.run([python_path] + "-m pip install binsync".split(" "))
    # just in case...
    location = subprocess.run(["which", "python3"], stdout=subprocess.PIPE)
    python_path = location.stdout.strip()
    subprocess.run([python_path.decode()] + "-m pip install git+https://github.com/angr/binsync".split(" "))


def install():
    # confirm install platform works
    platform = find_platform()
    if platform not in [PlatformType.LINUX, PlatformType.MACOS]:
        os.remove("oneliner.py")
        raise Exception("Platform is not supported for oneliner install.")

    # find python executable
    if platform == PlatformType.LINUX:
        python_path = sys.executable
    elif platform == PlatformType.MACOS:
        for lib_path in sys.path:
            basename = os.path.basename(lib_path)
            if basename.startswith("python") and not basename.endswith(".zip"):
                python_path = os.path.join(lib_path, f"../../bin/{basename}")
                if os.path.exists(python_path):
                    break
        else:
            os.remove("oneliner.py")
            raise Exception("Unable to locate your local python executable. Please use manual install.")

    pip_install_binsync(python_path)
    print("[+] Successfully installed BinSync to IDA pip")

    # find plugin path
    not_found = False
    for plugin_path in sys.path:
        if os.path.basename(plugin_path) == "plugins" and "ida" in plugin_path:
            break
    else:
        not_found = True

    # tray again with a less reliable search
    if not_found:
        for plugin_path in sys.path:
            if os.path.basename(plugin_path) == "python" and "ida" in plugin_path:
                break
        else:
            os.remove("oneliner.py")
            raise Exception("Unable to find the local plugins folder to install BinSync, install manually!")

        plugin_path = os.path.join(os.path.dirname(plugin_path), "plugins")

    plugin_install_binsync(plugin_path)
    print("[+] Successfully installed BinSync IDA Plugin into the plugins folder")
    print("[+] Install finished. PLEASE RESTART IDA for plugin to be loaded")
    os.remove("oneliner.py")



# ----------------------------------------------------------------------------
# This file contains the BinSyncController class which acts as the the
# bridge between the plugin UI and direct calls to the binsync client found in
# the core of binsync. In the controller, you will find code used to make
# pushes and pulls of user changes.
#
# You will also notice that the BinSyncController runs two extra threads in
# it:
#   1. BinSync "git pulling" thread to constantly get changes from others
#   2. Command Routine to get hooked changes to IDA attributes
#
# The second point is more complicated because it acts as the queue of
# runnable actions that are queued from inside the hooks.py file.
# Essentially, every change that happens in IDA from the main user triggers
# a hook which will push an action to be preformed onto the command queue;
# Causing a "git push" on every change.
#
# ----------------------------------------------------------------------------

import re
import threading
import functools
from typing import Dict, List, Tuple, Optional, Iterable, Any
import hashlib
import logging

from binaryninja import SymbolType
from binaryninjaui import (
    UIContext,
    DockHandler,
    DockContextHandler,
    UIAction,
    UIActionHandler,
    Menu,
)
from binaryninja.enums import MessageBoxButtonSet, MessageBoxIcon, VariableSourceType
from binaryninja.mainthread import execute_on_main_thread, is_main_thread

from binsync.common.controller import BinSyncController, fill_event, init_checker
from binsync.data import (
    State, User, Artifact,
    Function, FunctionHeader, FunctionArgument, StackVariable, StackOffsetType,
    Comment, GlobalVariable, Patch,
    Enum, Struct
)
import binsync

from .artifact_lifter import BinjaArtifactLifter

l = logging.getLogger(__name__)

#
# Helpers
#


def background_and_wait(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        output = [None]

        def thunk():
            output[0] = func(*args, **kwargs)
            return 1

        thread = threading.Thread(target=thunk)
        thread.start()
        thread.join()

        return output[0]
    return wrapper


#
# Controller
#

class BinjaBinSyncController(BinSyncController):
    def __init__(self):
        super(BinjaBinSyncController, self).__init__(artifact_lifter=BinjaArtifactLifter(self))
        self.bv = None

    def binary_hash(self) -> str:
        hash_ = ""
        try:
            hash_ = hashlib.md5(self.bv.file.raw[:]).hexdigest()
        except Exception:
            pass

        return hash_

    def active_context(self):
        all_contexts = UIContext.allContexts()
        if not all_contexts:
            return None

        ctx = all_contexts[0]
        handler = ctx.contentActionHandler()
        if handler is None:
            return None

        actionContext = handler.actionContext()
        func = actionContext.function
        if func is None:
            return None

        return binsync.data.Function(
            func.start, 0, header=FunctionHeader(func.name, func.start)
        )

    def binary_path(self) -> Optional[str]:
        try:
            return self.bv.file.filename
        except Exception:
            return None

    def get_func_size(self, func_addr) -> int:
        func = self.bv.get_function_at(func_addr)
        if not func:
            return 0

        return func.highest_address - func.start

    def goto_address(self, func_addr) -> None:
        self.bv.offset = func_addr

    #
    # Fillers
    #

    @init_checker
    def fill_struct(self, struct_name, header=True, members=True, artifact=None, **kwargs):
        return False

    @init_checker
    def fill_global_var(self, var_addr, user=None, artifact=None, **kwargs):
        return False

    @init_checker
    @background_and_wait
    @fill_event
    def fill_function(self, func_addr, user=None, artifact=None, **kwargs):
        """
        Grab all relevant information from the specified user and fill the @bn_func.
        """
        sync_func: Function = artifact
        bn_func = self.bv.get_function_at(sync_func.addr)

        changes = super(BinjaBinSyncController, self).fill_function(
            func_addr, user=user, artifact=artifact, bn_func=bn_func, **kwargs
        )
        bn_func.reanalyze()
        return changes

    @init_checker
    @fill_event
    def fill_function_header(self, func_addr, user=None, artifact=None, bn_func=None, **kwargs):
        updates = False
        sync_header: FunctionHeader = artifact

        if sync_header:
            # func name
            if sync_header.name and sync_header.name != bn_func.name:
                bn_func.name = sync_header.name
                updates |= True

            # ret type
            if sync_header.ret_type and \
                    sync_header.ret_type != bn_func.return_type.get_string_before_name():

                valid_type = False
                try:
                    new_type, _ = self.bv.parse_type_string(sync_header.ret_type)
                    valid_type = True
                except Exception:
                    pass

                if valid_type:
                    bn_func.return_type = new_type
                    updates |= True

            # parameters
            if sync_header.args:
                prototype_tokens = [sync_header.ret_type] if sync_header.ret_type \
                    else [bn_func.return_type.get_string_before_name()]

                prototype_tokens.append("(")
                for idx, func_arg in sync_header.args.items():
                    prototype_tokens.append(func_arg.type_str)
                    prototype_tokens.append(func_arg.name)
                    prototype_tokens.append(",")

                if prototype_tokens[-1] == ",":
                    prototype_tokens[-1] = ")"

                prototype_str = " ".join(prototype_tokens)

                valid_type = False
                try:
                    bn_prototype, _ = self.bv.parse_type_string(prototype_str)
                    valid_type = True
                except Exception:
                    pass

                if valid_type:
                    bn_func.function_type = bn_prototype
                    updates |= True

        return updates

    @init_checker
    @fill_event
    def fill_stack_variable(self, func_addr, offset, user=None, artifact=None, bn_func=None, **kwargs):
        updates = False
        bs_stack_var: StackVariable = artifact

        existing_stack_vars: Dict[int, Any] = {
            v.storage: v for v in bn_func.stack_layout
            if v.source_type == VariableSourceType.StackVariableSourceType
        }

        bn_offset = bs_stack_var.get_offset(StackOffsetType.BINJA)

        if bn_offset in existing_stack_vars:
            if existing_stack_vars[bn_offset].name != bs_stack_var.name:
                existing_stack_vars[bn_offset].name = bs_stack_var.name

            valid_type = False
            try:
                type_, _ = self.bv.parse_type_string(bs_stack_var.type)
                valid_type = True
            except Exception:
                pass

            if valid_type:
                if existing_stack_vars[bn_offset].type != type_:
                    existing_stack_vars[bn_offset].type = type_
                try:
                    bn_func.create_user_stack_var(bn_offset, type_, bs_stack_var.name)
                    bn_func.create_auto_stack_var(bn_offset, type_, bs_stack_var.name)
                except Exception as e:
                    l.warning(f"BinSync could not sync stack variable at offset {bn_offset}: {e}")

                updates |= True

        return updates

    @init_checker
    @fill_event
    def fill_comment(self, addr, user=None, artifact=None, bn_func=None, **kwargs):
        # TODO: check if the comment changed when set!
        comment: Comment = artifact
        bn_func.set_comment_at(comment.addr, comment.comment)

        return True

    #
    # Artifact API
    #

    def functions(self) -> Dict[int, Function]:
        funcs = {}
        for bn_func in self.bv.functions:
            if bn_func.symbol.type != SymbolType.FunctionSymbol:
                continue

            funcs[bn_func.start] = Function(bn_func.start, bn_func.total_bytes)
            funcs[bn_func.start].name = bn_func.name

        return funcs

    def function(self, addr) -> Optional[Function]:
        """
        TODO: fix how types and offsets are set

        @param addr:
        @return:
        """
        bn_func = self.bv.get_function_at(addr)
        if not bn_func:
            return None

        func = Function(bn_func.start, bn_func.total_bytes)
        func_header = FunctionHeader(
            bn_func.name,
            func.addr,
            ret_type=bn_func.return_type.get_string_before_name(),
            args={
                idx: FunctionArgument(idx, param.name, str(param.type), param.type.width)
                for idx, param in enumerate(bn_func.function_type.parameters)
            }
        )
        stack_vars = {
            v.storage: StackVariable(v.storage, StackOffsetType.BINJA, v.name, str(v.type), v.type.width, func.addr)
            for v in bn_func.stack_layout if v.source_type == VariableSourceType.StackVariableSourceType
        }
        func.header = func_header
        func.stack_vars = stack_vars

        return func

    def global_vars(self) -> Dict[int, GlobalVariable]:
        return {
            addr: GlobalVariable(addr, self.bv.get_symbol_at(addr) or f"data_{addr:x}")
            for addr, var in self.bv.data_vars.items()
        }
    
    def global_var(self, addr) -> Optional[GlobalVariable]:
        try:
            var = self.bv.data_vars[addr]
        except KeyError:
            return None 
            
        gvar = GlobalVariable(
            addr, self.bv.get_symbol_at(addr) or f"data_{addr:x}", type_str=str(var.type), size=var.type.width
        )
        return gvar


import logging
import os
import pathlib
import time
from functools import wraps
from typing import Dict, Iterable, Optional, Union, List

import git
import toml
from sortedcontainers import SortedDict

from binsync.data import (
    Comment,
    Enum,
    Function,
    FunctionHeader,
    GlobalVariable,
    Patch,
    StackVariable,
    Struct
)
from binsync.core.errors import MetadataNotFoundError


l = logging.getLogger(__name__)


class ArtifactType:
    UNSET = None
    FUNCTION = "function"
    STRUCT = "struct"
    PATCH = "patch"
    COMMENT = "comment"
    GLOBAL_VAR = "global variable"
    ENUM = "enum"


#
# Helper Funcs
#

def dirty_checker(f):
    @wraps(f)
    def dirtycheck(self, *args, **kwargs):
        r = f(self, *args, **kwargs)
        if r is True:
            self._dirty = True
        return r

    return dirtycheck


def update_last_change(f):
    @wraps(f)
    def _update_last_change(self, *args, **kwargs):
        should_set = kwargs.pop('set_last_change', True)
        artifact = args[0]

        # make a function if one does not exist
        if isinstance(artifact, (FunctionHeader, StackVariable)):
            func = self.get_or_make_function(artifact.addr)

        if not should_set:
            return f(self, *args, **kwargs)
        artifact.last_change = int(time.time())

        # Comment
        if isinstance(artifact, Comment):
            artifact_loc = artifact.addr
            artifact_type = ArtifactType.COMMENT
            # update function its in, if it's in a function
            func = self.find_func_for_addr(artifact.addr)
            if func:
                func.last_change = artifact.last_change

        # Stack Var
        elif isinstance(artifact, StackVariable):
            artifact_loc = artifact.addr
            artifact_type = ArtifactType.FUNCTION
            func.last_change = artifact.last_change

        elif isinstance(artifact, Function):
            artifact_loc = artifact.addr
            artifact_type = ArtifactType.FUNCTION

        # Function Header
        elif isinstance(artifact, FunctionHeader):
            artifact_loc = artifact.addr
            artifact_type = ArtifactType.FUNCTION
            func.last_change = artifact.last_change

        # Patch
        elif isinstance(artifact, Patch):
            artifact_loc = artifact.offset
            artifact_type = ArtifactType.PATCH

        # Struct
        elif isinstance(artifact, Struct):
            artifact_loc = artifact.name
            artifact_type = ArtifactType.STRUCT

        # Global Var
        elif isinstance(artifact, GlobalVariable):
            artifact_loc = artifact.addr
            artifact_type = ArtifactType.GLOBAL_VAR

        # Enum
        elif isinstance(artifact, Enum):
            artifact_loc = artifact.name
            artifact_type = ArtifactType.ENUM

        else:
            raise Exception("Undefined Artifact Type!")

        self.last_push_artifact = artifact_loc
        self.last_push_time = artifact.last_change
        self.last_push_artifact_type = artifact_type

        return f(self, *args, **kwargs)

    return _update_last_change


def list_files_in_dir(src: Union[pathlib.Path, git.Tree], dir_name, client=None) -> List[str]:
    if client and isinstance(src, git.Tree):
        files = client.list_files_in_tree(src)
        return [name for name in files if name.startswith(dir_name)]

    # load from filesystem
    if not src:
        src = pathlib.Path("../core")

    if not src.joinpath(pathlib.Path(dir_name)).exists():
        return []

    dir_name_path = pathlib.Path(dir_name)
    dir_path = src.joinpath(dir_name)
    return [
        str(dir_name_path.joinpath(pathlib.Path(name))) for name in os.listdir(dir_path)
    ]


def load_toml_from_file(src: Union[pathlib.Path, git.Tree], filename, client=None):
    if client and isinstance(src, git.Tree):
        file_data = client.load_file_from_tree(src, filename)
    else:
        if not src:
            src = pathlib.Path("../core")

        with open(src.joinpath(filename), "r") as fp:
            file_data = fp.read()

    return toml.loads(file_data)


#
# State Defn & Operators
#

class State:
    """
    The state.

    :ivar str user:     Name of the user.
    :ivar int version:  Version of the state, starting from 0.
    """

    def __init__(self, user, version=None, client=None):
        # metadata info
        self.user = user  # type: str
        self.version = version if version is not None else 0  # type: int
        self.last_push_artifact = -1
        self.last_push_artifact_type = -1
        self.last_push_time = -1

        # the client
        self.client = client  # type: Optional[Client]

        # data
        self.functions: Dict[int, Function] = {}
        self.comments: Dict[int, Comment] = {}
        self.structs: Dict[str, Struct] = {}
        self.patches: Dict[int, Patch] = SortedDict()
        self.global_vars: Dict[int, GlobalVariable] = {}
        self.enums: Dict[str, Enum] = {}

        # state is dirty on creation (metadata)
        self._dirty = True  # type: bool

    def __eq__(self, other):
        if isinstance(other, State):
            return other.functions == self.functions \
                   and other.comments == self.comments \
                   and other.structs == self.structs \
                   and other.patches == self.patches \
                   and other.global_vars == self.global_vars \
                   and other.enums == self.enums
        return False

    def copy(self):
        state = State(self.user, version=self.version, client=self.client)
        state._dirty = False
        artifacts = ["functions", "comments", "structs", "patches", "global_vars", "enums"]
        for artifact in artifacts:
            setattr(
                state,
                artifact,
                {k: v.copy() for k, v in getattr(self, artifact).items()}
            )

        return state

    def __str__(self):
        return f"<State: {self.user} " \
               f"funcs={len(self.functions)} " \
               f"cmts={len(self.comments)} " \
               f"globals={len(self.structs) + len(self.global_vars) + len(self.enums)}" \
               f">"

    def __repr__(self):
        return self.__str__()

    @property
    def dirty(self):
        return self._dirty

    def _dump_data(self, dst: Union[pathlib.Path, git.IndexFile], filename, data):
        # dump using Git files
        if self.client and isinstance(dst, git.IndexFile):
            self.client.add_data(dst, filename, data)
            return

        # dump using filesystem
        if not dst:
            dst = pathlib.Path("../core")

        out_path = dst.joinpath(filename)
        pathlib.Path(out_path).parent.mkdir(parents=True, exist_ok=True)
        with open(out_path, "wb") as fp:
            fp.write(data)

    def dump_metadata(self, dst: Union[pathlib.Path, git.IndexFile]):
        d = {
            "user": self.user,
            "version": self.version,
            "last_push_time": self.last_push_time,
            "last_push_artifact": self.last_push_artifact,
            "last_push_artifact_type": self.last_push_artifact_type,
        }
        self._dump_data(dst, 'metadata.toml', toml.dumps(d).encode())

    def dump(self, dst: Union[pathlib.Path, git.IndexFile]):
        if isinstance(dst, str):
            dst = pathlib.Path(dst)

        # dump metadata
        self.dump_metadata(dst)

        # dump functions, one file per function in ./functions/
        for addr, func in self.functions.items():
            path = pathlib.Path('functions').joinpath("%08x.toml" % addr)
            self._dump_data(dst, path, func.dump().encode())

        # dump structs, one file per struct in ./structs/
        for s_name, struct in self.structs.items():
            path = pathlib.Path('structs').joinpath(f"{s_name}.toml")
            self._dump_data(dst, path, struct.dump().encode())

        # dump comments
        self._dump_data(dst, 'comments.toml', toml.dumps(Comment.dump_many(self.comments)).encode())

        # dump patches
        self._dump_data(dst, 'patches.toml', toml.dumps(Patch.dump_many(self.patches)).encode())

        # dump global vars
        self._dump_data(dst, 'global_vars.toml', toml.dumps(GlobalVariable.dump_many(self.global_vars)).encode())

        # dump enums
        self._dump_data(dst, 'enums.toml', toml.dumps(Enum.dump_many(self.enums)).encode())

    @classmethod
    def parse(cls, src: Union[pathlib.Path, git.Tree], version=None, client=None):
        if isinstance(src, str):
            src = pathlib.Path(src)

        state = cls(None, version=version, client=client)

        # load metadata
        try:
            metadata = load_toml_from_file(src, "metadata.toml", client=client)
        except:
            # metadata is not found
            raise MetadataNotFoundError()
        state.user = metadata["user"]
        state.version = version if version is not None else metadata["version"]

        # load functions
        function_files = list_files_in_dir(src, "functions", client=client)
        for func_file in function_files:
            try:
                func_toml = load_toml_from_file(src, func_file, client=client)
            except:
                pass
            else:
                func = Function.load(func_toml)
                state.functions[func.addr] = func

        # load comments
        try:
            comments_toml = load_toml_from_file(src, "comments.toml", client=client)
        except:
            pass
        else:
            comments = {}
            for comment in Comment.load_many(comments_toml):
                comments[comment.addr] = comment
            state.comments = comments

        # load patches
        try:
            patches_toml = load_toml_from_file(src, "patches.toml", client=client)
        except:
            pass
        else:
            patches = {}
            for patch in Patch.load_many(patches_toml):
                patches[patch.offset] = patch
            state.patches = SortedDict(patches)

        # load global_vars
        try:
            global_vars_toml = load_toml_from_file(src, "global_vars.toml", client=client)
        except:
            pass
        else:
            global_vars = {}
            for global_var in GlobalVariable.load_many(global_vars_toml):
                global_vars[global_var.addr] = global_var
            state.global_vars = SortedDict(global_vars)

        # load enums
        try:
            enums_toml = load_toml_from_file(src, "enums.toml", client=client)
        except:
            pass
        else:
            state.enums = {
                enum.name: enum for enum in Enum.load_many(enums_toml)
            }

        # load structs
        struct_files = list_files_in_dir(src, "structs", client=client)
        for struct_file in struct_files:
            try:
                struct_toml = load_toml_from_file(src, struct_file, client=client)
            except:
                pass
            else:
                struct = Struct.load(struct_toml)
                state.structs[struct.name] = struct

        # clear the dirty bit
        state._dirty = False
        return state

    #
    # Setters
    #

    @dirty_checker
    @update_last_change
    def set_function(self, function: Function, set_last_change=True):
        if function.addr in self.functions and self.functions[function.addr] == function:
            return False

        self.functions[function.addr] = function
        return True

    @dirty_checker
    @update_last_change
    def set_function_header(self, func_header: FunctionHeader, set_last_change=True):
        if func_header.addr in self.functions and self.functions[func_header.addr] == func_header:
            return False

        self.functions[func_header.addr].header = func_header
        return True

    @dirty_checker
    @update_last_change
    def set_comment(self, comment: Comment, set_last_change=True):
        if not comment:
            return False

        try:
            old_cmt = self.comments[comment.addr]
        except KeyError:
            old_cmt = None

        if old_cmt != comment:
            self.comments[comment.addr] = comment
            return True

        return False

    @dirty_checker
    @update_last_change
    def set_patch(self, patch, addr, set_last_change=True):
        if not patch:
            return False

        try:
            old_patch = self.patches[addr]
        except KeyError:
            old_patch = None

        if old_patch != patch:
            self.patches[addr] = patch
            return True

        return False

    @dirty_checker
    @update_last_change
    def set_stack_variable(self, variable: StackVariable, set_last_change=True):
        if not variable:
            return False

        func = self.get_function(variable.addr)
        if not func:
            return False

        try:
            old_var = func.stack_vars[variable.stack_offset]
        except KeyError:
            old_var = None

        if old_var != variable:
            func.stack_vars[variable.stack_offset] = variable
            return True

        return False

    @dirty_checker
    @update_last_change
    def set_struct(self, struct: Struct, old_name=None, set_last_change=True):
        """
        Sets a struct in the current state. If old_name is not defined (None), then
        this indicates that the struct has not changed names. In that case, simply overwrite the
        internal representation of the struct.

        If the old_name is defined, than a struct has changed names. In that case, delete
        the internal struct data and delete the related .toml file.

        @param struct:
        @param old_name:
        @param set_last_change:
        @return:
        """
        if struct.name in self.structs \
                and self.structs[struct.name] == struct:
            # no updated is required
            return False

        # delete old struct only when we know what it is
        if old_name is not None:
            try:
                del self.structs[old_name]
                #remove_data(self.client.repo.index, os.path.join('structs', f'{old_name}.toml'))
            except KeyError:
                pass

        # set the new struct
        if struct.name is not None:
            self.structs[struct.name] = struct
            return True

        return False

    @dirty_checker
    @update_last_change
    def set_global_var(self, gloabl_var: GlobalVariable, set_last_change=True):
        try:
            old_gvar = self.global_vars[gloabl_var.addr]
        except KeyError:
            old_gvar = None

        if old_gvar != gloabl_var:
            self.global_vars[gloabl_var.addr] = gloabl_var
            return True

        return False

    @dirty_checker
    @update_last_change
    def set_enum(self, enum: Enum, set_last_change=True):
        try:
            old_enum = self.enums[enum.name]
        except KeyError:
            old_enum = None

        if old_enum != enum:
            self.enums[enum.name] = enum
            return True

        return False

    #
    # Getters
    #

    def get_or_make_function(self, addr) -> Function:
        try:
            func = self.functions[addr]
        except KeyError:
            self.functions[addr] = Function(addr, 0)
            func = self.functions[addr]

        return func

    def get_function(self, addr) -> Function:
        return self.functions.get(addr, None)

    def get_functions(self) -> Dict[int, Function]:
        return self.functions

    def get_function_header(self, addr) -> Optional[FunctionHeader]:
        func = self.get_function(addr)
        if not func:
            return None

        return func.header

    def get_function_headers(self) -> Dict[int, FunctionHeader]:
        return {
            addr: func.header
            for addr, func in self.functions.items() if func.header
        }

    def get_comment(self, addr) -> Comment:
        try:
            cmt = self.comments[addr]
        except KeyError:
            cmt = None

        return cmt

    def get_func_comments(self, func_addr) -> Dict[int, Comment]:
        try:
            func = self.functions[func_addr]
        except KeyError:
            return {}

        return {
            addr: cmt for addr, cmt in self.comments.items()
            if func.addr <= addr <= func.addr + func.size
        }

    def get_patch(self, addr) -> Patch:
        try:
            patch = self.patches[addr]
        except KeyError:
            patch = None

        return patch

    def get_patches(self) -> Dict[int, Patch]:
        return self.patches

    def get_stack_variable(self, func_addr, offset) -> Optional[StackVariable]:
        func = self.get_function(func_addr)
        if not func:
            return None

        try:
            stack_var = func.stack_vars[offset]
        except KeyError:
            stack_var = None

        return stack_var

    def get_stack_variables(self, func_addr) -> Dict[int, StackVariable]:
        func = self.get_function(func_addr)
        if not func:
            return {}

        return func.stack_vars

    def get_struct(self, struct_name) -> Optional[Struct]:
        try:
            struct = self.structs[struct_name]
        except KeyError:
            struct = None

        return struct

    def get_structs(self) -> Dict[str, Struct]:
        return self.structs

    def get_global_var(self, addr):
        try:
            gvar = self.global_vars[addr]
        except KeyError:
            gvar = None

        return gvar

    def get_global_vars(self):
        return self.global_vars

    def get_enum(self, name):
        try:
            enum = self.enums[name]
        except KeyError:
            enum = None

        return enum

    def get_enums(self):
        return self.enums

    def get_last_push_for_artifact_type(self, artifact_type):
        last_change = -1
        artifact = None

        if artifact_type == ArtifactType.FUNCTION:
            for function in self.functions.values():
                if function.last_change > last_change:
                    last_change = function.last_change
                    artifact = function.addr
        elif artifact_type == ArtifactType.STRUCT:
            for struct in self.structs.values():
                if struct.last_change > last_change:
                    last_change = struct.last_change
                    artifact = struct.name
        elif artifact_type == ArtifactType.PATCH:
            for patch in self.patches.values():
                if patch.last_change > last_change:
                    last_change = patch.last_change
                    artifact = patch.offset

        return tuple((artifact, last_change))

    #
    # Utils
    #

    def diff_comments(self, other_comments: Dict[int, Comment], diff_range=None):
        """

        :param other_comments:
        :param diff_range: [start_addr, end_addr]
        :return:
        """
        diff_dict = {}

        for addr, cmt in self.comments.items():
            # if addr is less than start or bigger than end
            if diff_range and (diff_range[0] > addr or addr >= diff_range[1]):
                continue

            try:
                other_cmt = other_comments[addr]
            except KeyError:
                other_cmt = None

            diff_dict[addr] = cmt.diff(other_cmt)

        for addr, cmt in other_comments.items():
            if diff_range and (diff_range[0] > addr or addr >= diff_range[1]):
                continue

            if addr in diff_dict:
                continue

            diff_dict[addr] = cmt.diff(None)

        return diff_dict

    def find_func_for_addr(self, search_addr):
        for func_addr, func in self.functions.items():
            if func.addr <= search_addr < (func.addr + func.size):
                return func
        else:
            return None



import logging

from binsync.data.artifact import Artifact
from binsync.data.type_parser import BSTypeParser, BSType

_l = logging.getLogger(name=__name__)


class ArtifactLifter:
    def __init__(self, controller, types=None):
        self.controller = controller
        self.type_parser = BSTypeParser(extra_types=types)

    #
    # Public API
    #

    def lift(self, artifact: Artifact):
        return self._lift_or_lower_artifact(artifact, "lift")

    def lower(self, artifact: Artifact):
        return self._lift_or_lower_artifact(artifact, "lower")

    #
    # TODO: deprecate these and standardize property names
    #

    def lift_ret_type(self, type_str):
        return self.lift_type(type_str)

    def lower_ret_type(self, type_str):
        return self.lower_type(type_str)

    def lift_type_str(self, type_str):
        return self.lift_type(type_str)

    def lower_type_str(self, type_str):
        return self.lower_type(type_str)

    #
    # Override Mandatory Funcs
    #

    def lift_type(self, type_str: str) -> str:
        #raise NotImplementedError
        print("lift type called")

    def lift_addr(self, addr: int) -> int:
        #raise NotImplementedError
        print("lift addr called")

    def lift_stack_offset(self, offset: int, func_addr: int) -> int:
        #raise NotImplementedError
        print("lift stack off called")

    def lower_type(self, type_str: str) -> str:
        #raise NotImplementedError
        print("lower type called")

    def lower_addr(self, addr: int) -> int:
        #raise NotImplementedError
        print("lower addr called")

    def lower_stack_offset(self, offset: int, func_addr: int) -> int:
        #raise NotImplementedError
        print("lower stack off called")

    #
    # Private
    #

    def _lift_or_lower_artifact(self, artifact, mode):
        target_attrs = ("ret_type", "type_str", "type", "stack_offset", "addr")
        if mode not in ("lower", "lift"):
            return None

        if not isinstance(artifact, Artifact):
            return artifact
        lifted_art = artifact.copy()

        # correct simple properties in the artifact
        for attr in target_attrs:
            if hasattr(lifted_art, attr):
                curr_val = getattr(lifted_art, attr)
                if not curr_val:
                    continue

                lifting_func = getattr(self, f"{mode}_{attr}")

                if attr == "stack_offset":
                    setattr(lifted_art, attr, lifting_func(curr_val, lifted_art.addr))
                else:
                    setattr(lifted_art, attr, lifting_func(curr_val))

        # recursively correct nested artifacts
        for attr in lifted_art.__slots__:
            attr_val = getattr(lifted_art, attr)
            if not attr_val:
                continue

            # nested function headers
            if attr == "header":
                setattr(lifted_art, attr, self._lift_or_lower_artifact(attr_val, mode))
            # nested args, stack_vars, or struct_members
            elif isinstance(attr_val, dict):
                nested_arts = {
                    k: self._lift_or_lower_artifact(v, mode) for k, v in attr_val.items()
                }
                setattr(lifted_art, attr, nested_arts)

        return lifted_art




import logging
from typing import Dict

from binsync.common.controller import BinSyncController
from binsync.common.ui.qt_objects import (
    QAbstractItemView,
    QHeaderView,
    QMenu,
    Qt,
    QTableWidget,
    QTableWidgetItem,
)
from binsync.common.ui.utils import QNumericItem, friendly_datetime
from binsync.data import Function
from binsync.data.state import State
from binsync.core.scheduler import SchedSpeed

l = logging.getLogger(__name__)

class QFunctionItem:
    def __init__(self, addr, name, user, last_push):
        self.addr = addr
        self.name = name
        self.user = user
        self.last_push = last_push

    def __eq__(self, other):
        return self.addr == other.addr

    def __hash__(self):
        return self.addr

    def widgets(self):
        # sort by int value
        addr = QNumericItem(hex(self.addr))
        addr.setData(Qt.UserRole, self.addr)

        name = QTableWidgetItem(self.name)
        user = QTableWidgetItem(self.user)

        # sort by unix value
        last_push = QNumericItem(friendly_datetime(self.last_push))
        last_push.setData(Qt.UserRole, self.last_push)

        widgets = [
            addr,
            name,
            user,
            last_push
        ]

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QFunctionTable(QTableWidget):

    HEADER = [
        'Addr',
        'Remote Name',
        'User',
        'Last Push'
    ]

    def __init__(self, controller: BinSyncController, parent=None):
        super(QFunctionTable, self).__init__(parent)
        self.controller = controller
        self.items = dict()

        self.setColumnCount(len(self.HEADER))
        self.setHorizontalHeaderLabels(self.HEADER)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.horizontalHeader().setHorizontalScrollMode(self.ScrollPerPixel)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignHCenter | Qt.Alignment(Qt.TextWordWrap))
        self.horizontalHeader().setMinimumWidth(160)
        self.setHorizontalScrollMode(self.ScrollPerPixel)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(24)

        self.setSortingEnabled(True)

        self.doubleClicked.connect(self._doubleclick_handler)
        self.last_table = set()

    def reload(self):
        self.setSortingEnabled(False)
        self.setRowCount(len(self.items))
        new_table = set(self.items.values())
        new_entries = new_table.difference(self.last_table)

        for idx, item in enumerate(new_entries):
            for i, attr in enumerate(item.widgets()):
                self.setItem(idx, i, attr)

        self.last_table = new_table
        self.viewport().update()
        self.setSortingEnabled(True)

    def contextMenuEvent(self, event):
        menu = QMenu(self)
        menu.setObjectName("binsync_function_table_context_menu")

        selected_row = self.rowAt(event.pos().y())
        item = self.item(selected_row, 0)
        if item is None:
            return
        func_addr = item.data(Qt.UserRole)
        
        menu.addAction("Sync", lambda: self.controller.fill_function(func_addr, user=self.item(selected_row, 2).text()))
        from_menu = menu.addMenu("Sync from...")
        
        for username in self._get_valid_users_for_func(func_addr):
            action = from_menu.addAction(username)
            action.triggered.connect(lambda chck, name=username: self.controller.fill_function(func_addr, user=name))  

        menu.popup(self.mapToGlobal(event.pos()))

    def update_table(self):
        # first check if any functions are unknown to the table
        for user in self.controller.users():
            state = self.controller.client.get_state(user=user.name)
            user_funcs: Dict[int, Function] = state.functions

            for func_addr, sync_func in user_funcs.items():
                # don't add functions that were never changed by the user
                if not sync_func.last_change:
                    continue

                func_change_time = sync_func.last_change
                # compare this users change time to the store change time
                if func_addr not in self.items or func_change_time >= self.items[func_addr].last_push:
                    remote_func_name = sync_func.name if sync_func.name else ""
                    self.items[func_addr] = QFunctionItem(func_addr, remote_func_name, user.name, func_change_time)

    def _get_valid_users_for_func(self, func_addr):
        for user in self.controller.users(priority=SchedSpeed.FAST):
            user_state: State = self.controller.client.get_state(user=user.name, priority=SchedSpeed.FAST)
            user_func = user_state.get_function(func_addr)

            # function must be changed by this user
            if not user_func or not user_func.last_change:
                continue

            yield user.name

    def _doubleclick_handler(self):
        # Doubleclick only allows for a single item select so just take first one from list
        row_idx = self.selectionModel().selectedIndexes()[0].row()
        row_addr = self.item(row_idx, 0).data(Qt.UserRole)
        self.controller.goto_address(row_addr)



from binsync.data.user import User
from binsync.data.artifact import Artifact
from binsync.data.func import Function, FunctionHeader, FunctionArgument
from binsync.data.comment import Comment
from binsync.data.patch import Patch
from binsync.data.stack_variable import StackVariable, StackOffsetType
from binsync.data.struct import StructMember, Struct
from binsync.data.global_variable import GlobalVariable
from binsync.data.enum import Enum
from binsync.data.state import State
from binsync.data.configuration import ProjectConfig, GlobalConfig
from binsync.data.type_parser import BSType, BSTypeParser



import logging
from typing import Dict

from binsync.common.controller import BinSyncController
from binsync.common.ui.qt_objects import (
    QAbstractItemView,
    QHeaderView,
    QMenu,
    Qt,
    QTableWidget,
    QTableWidgetItem,
)
from binsync.common.ui.utils import QNumericItem, friendly_datetime
from binsync.data import Function
from binsync.core.scheduler import SchedSpeed
from binsync.data.state import State

l = logging.getLogger(__name__)

class QActivityItem:
    def __init__(self, user, activity, last_push):
        self.user = user
        self.activity = activity
        self.last_push = last_push

    def widgets(self):
        if isinstance(self.activity, int):
            activity = QNumericItem(hex(self.activity))
            activity.setData(Qt.UserRole, self.activity)
        else:
            activity = QNumericItem(self.activity)
            # set to max number so its last
            activity.setData(Qt.UserRole, -1)

        user = QTableWidgetItem(self.user)

        # sort by unix value
        last_push = QNumericItem(friendly_datetime(self.last_push))
        last_push.setData(Qt.UserRole, self.last_push)

        widgets = [
            user,
            activity,
            last_push
        ]

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QActivityTable(QTableWidget):
    """
    The activity table shown in the Control Panel. This table is responsible for showing users information
    that is relevant to other users activity. The main user wants to know what others are doing, and how
    often they are doing it. This table should also allow users to sync from a user if they see them as
    being very active.

    TODO: refactor the below code to allow activity view to show any item, not just a function!
    """

    HEADER = [
        'User',
        'Activity',
        'Last Push'
    ]

    def __init__(self, controller: BinSyncController, parent=None):
        super(QActivityTable, self).__init__(parent)
        self.controller = controller
        self.items = []

        self.setColumnCount(len(self.HEADER))
        self.setHorizontalHeaderLabels(self.HEADER)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.horizontalHeader().setHorizontalScrollMode(self.ScrollPerPixel)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignHCenter | Qt.Alignment(Qt.TextWordWrap))
        self.horizontalHeader().setMinimumWidth(160)
        self.setHorizontalScrollMode(self.ScrollPerPixel)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(24)

        self.setSortingEnabled(True)

    def reload(self):
        self.setSortingEnabled(False)
        self.setRowCount(len(self.items))

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        self.viewport().update()
        self.setSortingEnabled(True)

    def contextMenuEvent(self, event):
        menu = QMenu(self)
        menu.setObjectName("binsync_activity_table_context_menu")

        selected_row = self.rowAt(event.pos().y())
        item = self.item(selected_row, 0)
        if item is None:
            return
        username = item.text()
        menu.addAction("Sync", lambda: self.controller.fill_function(self.item(selected_row, 1).data(Qt.UserRole), user=username))

        menu.addAction("Sync-All", lambda: self.controller.fill_all(user=username))

        for_menu = menu.addMenu("Sync for...")
        for func_addr_str in self._get_valid_funcs_for_user(username):
            action = for_menu.addAction(func_addr_str)
            action.triggered.connect(lambda chk, func=func_addr_str: self.controller.fill_function(int(func, 16), user=username))

        menu.popup(self.mapToGlobal(event.pos()))

    def update_table(self):
        self.items = []

        # first check if any functions are unknown to the table
        for user in self.controller.users():
            changed_funcs = {}
            state = self.controller.client.get_state(user=user.name)
            user_funcs: Dict[int, Function] = state.functions

            for func_addr, sync_func in user_funcs.items():
                func_change_time = sync_func.last_change

                # don't add functions that were never changed by the user
                if not sync_func.last_change:
                    continue

                # check if we already know about it
                if func_addr in changed_funcs:
                    # compare this users change time to the store change time
                    if not func_change_time or func_change_time < changed_funcs[func_addr]:
                        continue

                changed_funcs[func_addr] = func_change_time

            if len(changed_funcs) > 0:
                most_recent_func = list(changed_funcs)[0]
                last_state_change = state.last_push_time \
                    if not state.last_push_time \
                    else list(changed_funcs.values())[0]
            else:
                most_recent_func = ""
                last_state_change = state.last_push_time

            self.items.append(
                QActivityItem(user.name, most_recent_func, last_state_change)
            )

    def _get_valid_funcs_for_user(self, username):
        user_state: State = self.controller.client.get_state(user=username, priority=SchedSpeed.FAST)
        func_addrs = [addr for addr in user_state.functions]

        func_addrs.sort()
        for func_addr in func_addrs:
            yield hex(func_addr)



import logging
import os
import time
import pathlib
import traceback
from typing import Optional
import git

import toml
from binsync.core.client import ConnectionWarnings, BINSYNC_ROOT_BRANCH
from binsync.data.configuration import ProjectConfig
from binsync.common.ui.qt_objects import (
    QCheckBox,
    QDialog,
    QDir,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
)

l = logging.getLogger(__name__)

class SyncConfig(QDialog):
    """
    The dialog that allows a user to config a BinSync client for:
    - initing a local repo
    - cloning a remote
    - using a locally pulled remote repo
    """
    def __init__(self, controller, open_magic_sync=True, load_config=True, parent=None):
        super().__init__(parent)
        self.controller = controller
        self.open_magic_sync = open_magic_sync
        self.should_load_config = load_config
        self.setWindowTitle("Configure BinSync")

        self._main_layout = QVBoxLayout()
        self._user_edit = None  # type:QLineEdit
        self._repo_edit = None  # type:QLineEdit
        self._remote_edit = None  # type:QLineEdit
        self._initrepo_checkbox = None  # type:QCheckBox

        self._init_widgets()
        self.setLayout(self._main_layout)
        self.show()

    def _init_widgets(self):
        upper_layout = QGridLayout()

        # user label
        user_label = QLabel(self)
        user_label.setText("User name")
        user_label.setToolTip(
            "The name your user will be saved as on remote. Can be anything other that 'root'. This name does not "
            "need to be the same as your Git username."
        )

        self._user_edit = QLineEdit(self)

        row = 0
        upper_layout.addWidget(user_label, row, 0)
        upper_layout.addWidget(self._user_edit, row, 1)
        row += 1

        # binsync label
        binsync_label = QLabel(self)
        binsync_label.setText("Git repo")
        binsync_label.setToolTip(
            "The path to a locally cloned Git repo. This can be ignored if you have not cloned down a remote "
            "repo yet, but you have the URL. The local repo can also be an empty folder that will be turned into "
            "a BinSync database by selecting 'init_remote'."
        )

        # repo path
        self._repo_edit = QLineEdit(self)
        self._repo_edit.textChanged.connect(self._on_repo_textchanged)
        #self._repo_edit.setFixedWidth(150)

        # repo path selection button
        repo_button = QPushButton(self)
        repo_button.setText("...")
        repo_button.clicked.connect(self._on_repo_clicked)
        repo_button.setFixedWidth(40)

        upper_layout.addWidget(binsync_label, row, 0)
        upper_layout.addWidget(self._repo_edit, row, 1)
        upper_layout.addWidget(repo_button, row, 2)
        row += 1

        # clone from a remote URL
        self.remote_label = QLabel(self)
        self.remote_label.setText("Remote URL")
        self.remote_label.setToolTip(
            "The URL to a remove Git repo. This is not required if you have already selected a locally cloned "
            "Git repo that has a remote. This repo will be cloned to the same location as your binary if a local "
            "repo does not exist yet."
        )


        self._remote_edit = QLineEdit(self)

        upper_layout.addWidget(self.remote_label, row, 0)
        upper_layout.addWidget(self._remote_edit, row, 1)
        row += 1

        # initialize repo checkbox
        self._initrepo_checkbox = QCheckBox(self)
        self._initrepo_checkbox.setText("Init Remote")
        self._initrepo_checkbox.setToolTip(
            "Ether inits the local folder, making it into a Git repo, or updates the remote references of the current "
            "Git repo to have the correct BinSync layout which includes the root branch and the new user. Should only "
            "ever be used when first creating a BinSync repo for all users."
        )
        self._initrepo_checkbox.setChecked(False)
        #self._initrepo_checkbox.setEnabled(False)

        upper_layout.addWidget(self._initrepo_checkbox, row, 1)
        row += 1

        # buttons
        self._ok_button = QPushButton(self)
        self._ok_button.setText("OK")
        self._ok_button.setDefault(True)
        self._ok_button.clicked.connect(self._on_ok_clicked)

        cancel_button = QPushButton(self)
        cancel_button.setText("Cancel")
        cancel_button.clicked.connect(self._on_cancel_clicked)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self._ok_button)
        buttons_layout.addWidget(cancel_button)

        # main layout
        self._main_layout.addLayout(upper_layout)
        self._main_layout.addLayout(buttons_layout)

        # change the text if config exists
        if self.should_load_config:
            self.load_saved_config()

    #
    # Event handlers
    #

    def _on_ok_clicked(self):
        user = self._user_edit.text()
        path = self._repo_edit.text()
        remote_url = self._remote_edit.text()
        init_repo = self._initrepo_checkbox.isChecked()

        l.debug("Attempting to connect to/init repo, user: %s | path: %s | init_repo? %r", user, path, init_repo)

        if not user:
            QMessageBox(self).critical(None, "Invalid user name",
                                       "User name cannot be empty."
                                       )
            return

        if user.lower() == "__root__":
            QMessageBox(self).critical(None, "Invalid user name",
                                       "User name cannot (and should not) be \'__root__\'."
                                       )
            return

        if not remote_url and not os.path.isdir(path) and not init_repo:
            QMessageBox(self).critical(None, "Repo does not exist",
                                       "The specified sync directory does not exist. "
                                       "Do you maybe want to initialize it?"
                                       )
            return

        # convert to remote repo if no local is provided
        if self.is_git_repo(path):
            remote_url = None

        if remote_url and not path:
            path = os.path.join(os.path.dirname(self.controller.binary_path() or ""),
                                os.path.basename(self.controller.binary_path() or "") + "_bs"
                                )

        try:
            connection_warnings = self.controller.connect(user, path, init_repo=init_repo, remote_url=remote_url)
            pass
        except Exception as e:
            l.critical("Error connecting to specified repository!")
            QMessageBox(self).critical(None, "Error connecting to repository", str(e))
            traceback.print_exc()
            return

        #
        # controller is now successfully connected to a real BinSync client. Everything from this point
        # onwards assumes that all normal client properties and functions work.
        #

        # warn user of anything that might look off
        self._parse_and_display_connection_warnings(connection_warnings)
        l.info(f"Client has connected to sync repo with user: {user}.")

        # create and save config if possible
        saved_config = self.save_config()
        if saved_config:
            l.debug(f"Configuration file was saved to {saved_config}.")

        self.close()

    def _on_repo_clicked(self):
        if 'SNAP' in os.environ:
            directory = QFileDialog.getExistingDirectory(self, "Select sync repo", "",
                                                    QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks | QFileDialog.DontUseNativeDialog)
        else:
            directory = QFileDialog.getExistingDirectory(self, "Select sync repo", "",
                                                         QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks)
        self._repo_edit.setText(QDir.toNativeSeparators(directory))

    def _on_repo_textchanged(self, new_text):
        path = new_text
        if pathlib.Path(path).exists() and self.is_git_repo(path):
            repo = git.Repo(path)
            try:
                url = repo.remote().url
            except Exception:
                url = None

            self._remote_edit.setText(url or "")
            self._remote_edit.setEnabled(False)
            self.remote_label.setStyleSheet('color: gray')

            is_binsync_inited = False
            try:
                is_binsync_inited = any(ref.name.endswith(BINSYNC_ROOT_BRANCH) for ref in repo.remote().refs)
            except Exception:
                pass

            if is_binsync_inited:
                self._initrepo_checkbox.setChecked(False)
                self._initrepo_checkbox.setStyleSheet(
                    'QCheckBox::unchecked { color: gray } '
                    'QCheckBox::checked { color: gray }'
                )
                self._initrepo_checkbox.setEnabled(False)
            else:
                self._initrepo_checkbox.setEnabled(True)
                self._initrepo_checkbox.setStyleSheet(
                    'QCheckBox::unchecked { color: white } '
                    'QCheckBox::checked { color: white }'
                )
        else:
            self._remote_edit.setEnabled(True)
            self._initrepo_checkbox.setEnabled(True)
            self.remote_label.setStyleSheet('color: white')
            self._initrepo_checkbox.setStyleSheet(
                'QCheckBox::unchecked { color: white } '
                'QCheckBox::checked { color: white }'
            )

    def _on_cancel_clicked(self):
        self.close()

    #
    # Utils
    #

    def load_saved_config(self) -> bool:
        config = ProjectConfig.load_from_file(self.controller.binary_path() or "")
        if not config:
            return False

        user = config.user or ""
        repo = config.repo_path or ""
        remote = config.remote if config.remote and not config.repo_path else ""

        self._user_edit.setText(user)
        self._repo_edit.setText(repo)
        self._on_repo_textchanged(repo)
        self._remote_edit.setText(remote)
        return True

    def save_config(self) -> Optional[str]:
        user = self._user_edit.text()
        remote = self._remote_edit.text()
        repo = self._repo_edit.text()

        if remote and not repo:
            repo = str(pathlib.Path(self.controller.client.repo_root).absolute())

        config = ProjectConfig(
            self.controller.binary_path() or "",
            user=user,
            repo_path=repo,
            remote=remote
        )
        if not config:
            return config

        return config.save()


    #
    # Static methods
    #

    @staticmethod
    def is_git_repo(path):
        return os.path.isdir(os.path.join(path, ".git"))

    @staticmethod
    def _parse_and_display_connection_warnings(warnings):
        warning_text = ""

        for warning in warnings:
            if warning == ConnectionWarnings.HASH_MISMATCH:
                warning_text += "Warning: the hash stored for this BinSync project does not match " \
                                "the hash of the binary you are attempting to analyze. It's possible " \
                                "you are working on a different binary.\n"

        if len(warning_text) > 0:
            QMessageBox.warning(
                None,
                "BinSync: Connection Warnings",
                warning_text,
                QMessageBox.Ok,
            )



import logging
import sys

from binsync.common.ui.qt_objects import QVBoxLayout, QMainWindow, QApplication
from binsync.common.ui.control_panel import ControlPanel
from binsync.common.ui.config_dialog import SyncConfig

from .controller import GhidraBinSyncController


l = logging.getLogger(__name__)


class ControlPanelWindow(QMainWindow):
    """
    The class for the window that shows changes/info to BinSync data. This includes things like
    changes to functions or structs.
    """

    def __init__(self):
        super(ControlPanelWindow, self).__init__()
        self.setWindowTitle("BinSync")
        self.width_hint = 300

        self.controller: GhidraBinSyncController = GhidraBinSyncController()
        self.control_panel = ControlPanel(self.controller)

        self._init_widgets()

    #
    # Private methods
    #

    def _init_widgets(self):
        self.control_panel.show()
        self.setCentralWidget(self.control_panel)

    #
    # handlers
    #

    def configure(self):
        # setup bridge and alert it we are configuring
        bridge_connected = self.controller.connect_ghidra_client()

        config = SyncConfig(self.controller)
        config.show()
        config.exec_()
        client_connected = self.controller.check_client()

        if bridge_connected and client_connected:
            self.controller.ghidra.alert_ui_configured(client_connected)

        return client_connected and bridge_connected

    def closeEvent(self, event):
        self.controller.ghidra.server.stop()


def start_ui():
    app = QApplication()
    cp_window = ControlPanelWindow()

    # control panel should stay hidden until a good config happens
    cp_window.hide()
    connected = cp_window.configure()
    if connected:
        cp_window.show()
    else:
        sys.exit(1)

    app.exec_()



from collections import OrderedDict
from typing import Dict

import toml

from binsync.data.artifact import Artifact


class Enum(Artifact):
    __slots__ = Artifact.__slots__ + (
        "name",
        "members",
    )

    def __init__(self, name, members: Dict[str, int], last_change=None):
        super(Enum, self).__init__(last_change=last_change)
        self.name = name
        # sorts map by the int value
        self.members = OrderedDict(sorted(members.items(), key=lambda kv: kv[1]))

    def __str__(self):
        return f"<Enum: {self.name} member_count={len(self.members)}>"

    def __repr__(self):
        return self.__str__()

    @classmethod
    def parse(cls, s):
        en = Enum(None, {})
        en.__setstate__(toml.loads(s))
        return en

    @classmethod
    def load_many(cls, enums_toml):
        for enum_toml in enums_toml.values():
            enum = Enum(None, {})
            try:
                enum.__setstate__(enum_toml)
            except TypeError:
                # skip all incorrect ones
                continue
            yield enum

    @classmethod
    def dump_many(cls, enums):
        enums_ = {}

        for name, enum in enums.items():
            enums_[name] = enum.__getstate__()
        return enums_

    def copy(self):
        return Enum(
            self.name,
            self.members.copy(),
            last_change=self.last_change
        )

    def nonconflict_merge(self, enum2: "Enum", **kwargs):
        enum1: Enum = self.copy()
        if not enum2 or enum1 == enum2:
            return enum1.copy()

        master_state = kwargs.get("master_state", None)
        local_names = {mem for mem in enum1.members}
        if master_state:
            for _, enum in master_state.get_enums().items():
                local_names.union(set(enum.members.keys()))
        else:
            local_names = enum1.members

        constants = {
            value for value in enum1.members.values()
        }

        for name, constant in enum2.members.items():
            if name in local_names or constant in constants:
                continue
            enum1.members[name] = constant
        return enum1



import logging

from binsync.common import ArtifactLifter

l = logging.getLogger(name=__name__)


class GhidraArtifactLifter(ArtifactLifter):
    lift_map = {}

    def __init__(self, controller):
        super(GhidraArtifactLifter, self).__init__(controller)

    def lift_addr(self, addr: int) -> int:
        #return self.controller.rebase_addr(addr)
        return addr

    def lift_type(self, type_str: str) -> str:
        return type_str

    def lift_stack_offset(self, offset: int, func_addr: int) -> int:
        return offset

    def lower_addr(self, addr: int) -> int:
        #return self.controller.rebase_addr(addr, up=True)
        return addr

    def lower_type(self, type_str: str) -> str:
        return type_str

    def lower_stack_offset(self, offset: int, func_addr: int) -> int:
        return offset
